"""
This module provides the implementation of the commanding interfaces for the Alpha and Alpha+
controller using the new dynamic commanding scheme. The three main classes are the `AlphaPlusTelnetInterface`,
the `AlphaControllerInterface` and the `AlphaPlusControllerInterface`.

The `AlphaPlusTelnetInterface` directly talks to the device through the telnet protocol on port 23.

The `AlphaControllerInterface` provides an interface with methods that are compatible with both the Alpha controller
and the AlphaPlusController. This interface shall be sub-classed for proxy and controller classes that use the
alpha controller, like the PUNA Hexapod.

The `AlphaPlusControllerInterface` inherits from the `AlphaControllerInterface` and provides additional methods
that are specific for the alpha+ controllers. This class should be sub-classed for proxy and controller classes
that use the alpha+ controller, like the ZONDA hexapod.

"""
from __future__ import annotations

import logging
from functools import partial
from telnetlib import Telnet
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

from egse.device import DeviceConnectionError
from egse.device import DeviceConnectionInterface
from egse.device import DeviceInterface
from egse.device import DeviceTransport
from egse.mixin import add_cr_lf
from egse.mixin import dynamic_command
from egse.response import Failure
from egse.response import Success
from egse.settings import Settings
from egse.system import Timer
from egse.system import wait_until

LOGGER = logging.getLogger(__name__)
PUNA_PLUS = Settings.load("PUNA Alpha+ Controller")

# The following constants represent the index into the GENERAL_STATE list and are used in the code
# to match the name of a flag in the general_state.

HOME_COMPLETE = 6
IN_POSITION = 3
IN_MOTION = 4

GENERAL_STATE = [
    "Error",
    "System initialized",
    "Control on",
    "In position",
    "Motion task running",
    "Home task running",
    "Home complete",
    "Home virtual",
    "Phase found",
    "Brake on",
    "Motion restricted",
    "Power on encoders",
    "Power on limit switches",
    "Power on drives",
    "Emergency stop",
]


ACTUATOR_STATE = [
    "Error",
    "Control on",
    "In position",
    "Motion task running",
    "Home task running",
    "Home complete",
    "Phase found",
    "Brake on",
    "Home hardware input",
    "Negative hardware limit switch",
    "Positive hardware limit switch",
    "Software limit reached",
    "Following error",
    "Drive fault",
    "Encoder error",
]


ERROR_CODES = {
    1: "An emergency stop has been pressed.",
    2: "A safety input has been triggered. The status of the inputs is given in the DATA field.",
    3: "A temperature sensor has exceeded the limit threshold. Sensor number is given in DATA field.",
    4: "Controller system status error (Sys.Status).",
    5: "Controller ‘abort all’ input has been triggered. (Sys.AbortAll).",
    6: "Controller watchdog error (Sys.WDTFault).",
    7: "Configuration load error.",
    8: "Configuration failed: a wrong hexapod ID has been detected. Detected ID is given in DATA field.",
    9: "Home task has failed.",
    10: "Virtual home write task has failed.",
    11: "The motion program did not start in the defined timeout.",
    12: "The home task did not start in the defined timeout.",
    13: "A kinematic error has occurred. Kinematic error number is given in DATA field.",
    14: "Controller coordinate error status (Coord.ErrorStatus). Error number is given in DATA field.",
    15: "An error has been detected on encoder.",
    16: "Brake should have been engaged as the motor control was off.",
    17: "Controller motor status: Auxiliary fault (AuxFault).",
    18: "Controller motor status: Encoder loss (EncLoss).",
    19: "Controller motor status: Amplifier warning (AmpWarn).",
    20: "Controller motor status: Trigger not found (TriggerNotFound).",
    21: "Controller motor status: Integrated current 'I2T' fault (I2tFault).",
    22: "Controller motor status: Software positive limit reach (SoftPlusLimit).",
    23: "Controller motor status: Software negative limit reach (SoftMinusLimit).",
    24: "Controller motor status: Amplifier fault (AmpFault).",
    25: "Controller motor status: Stopped on hardware limit (LimitStop).",
    26: "Controller motor status: Fatal following error (FeFatal).",
    27: "Controller motor status: Warning following error (FeWarn).",
    28: "Controller motor status: Hardware positive limit reach (PlusLimit).",
    29: "Controller motor status: Hardware negative limit reach (MinusLimit).",
}


RETURN_CODES = {
    0: "Success.",
    -1: "Undefined error.",
    -10: "Wrong value for parameter at index 0.",
    -11: "Wrong value for parameter at index 1.",
    -12: "Wrong value for parameter at index 2.",
    -13: "Wrong value for parameter at index 3.",
    -14: "Wrong value for parameter at index 4.",
    -15: "Wrong value for parameter at index 5.",
    -16: "Wrong value for parameter at index 6.",
    -17: "Wrong value for parameter at index 7.",
    -18: "Wrong value for parameter at index 8.",
    -19: "Wrong value for parameter at index 9.",
    -20: "Wrong value for parameter at index 10.",
    -21: "Wrong value for parameter at index 11.",
    -22: "Wrong value for parameter at index 12.",
    -23: "Wrong value for parameter at index 13.",
    -24: "Wrong value for parameter at index 14.",
    -25: "Wrong value for parameter at index 15.",
    -26: "Wrong value for parameter at index 16.",
    -27: "Wrong value for parameter at index 17.",
    -28: "Wrong value for parameter at index 18.",
    -29: "Wrong value for parameter at index 19.",
    -30: "Unknown command number.",
    -31: "This configuration command is a 'get' only type.",
    -32: "This configuration command is a 'set' only type.",
    -33: "The axis number do not correspond to an axis defined on the controller.",
    -34: "A stop task is running.",
    -35: "All motors need to be control on.",
    -36: "All motors need to be control off.",
    -37: "Emergency stop is pressed.",
    -38: "A motion task is running.",
    -39: "A home task is running.",
    -40: "Requested move is not feasible.",
    -41: "Power supply of limit switches is off.",
    -42: "Power supply of encoders is off.",
    -43: "A fatal error is present. This type of error needs a controller restart to be removed.",
    -44: "An error is present, error reset is required.",
    -45: "Home is not completed.",
    -46: "Software option not available (can be linked to hardware configuration).",
    -47: "Virtual home: file was created on another controller (different MAC address).",
    -48: "Virtual home: some positions read in file are out of software limits.",
    -49: "Virtual home: file data were stored while hexapod was moving.",
    -50: "Virtual home: no data available.",
    -51: "Command has been rejected because another action is running.",
    -52: "Timeout waiting for home complete status.",
    -53: "Timeout waiting for control on status.",
    -54: "Timeout on motion program start.",
    -55: "Timeout on home task start.",
    -56: "Timeout on virtual home write file task.",
    -57: "Timeout on virtual home delete file task.",
    -58: "Timeout on virtual home read file task.",
    -59: "Timeout on disk access verification task.",
    -60: "Configuration file: save process failed.",
    -61: "Configuration file: loaded file is empty.",
    -62: "Configuration file: loaded data are corrupted.",
    -63: "No access to the memory disk.",
    -64: "File does not exist.",
    -65: "Folder access failed.",
    -66: "Creation of folder tree on the memory disk failed.",
    -67: "Generation or write of the checksum failed.",
    -68: "File read: no data or wrong data size.",
    -69: "File read: no checksum.",
    -70: "File read: incorrect checksum.",
    -71: "File write: failed.",
    -72: "File open: failed.",
    -73: "File delete: failed.",
    -74: "Get MAC address failed.",
    -75: "NaN (Not a Number) or infinite value found.",
    -76: "The coordinate system transformations are not initialized.",
    -77: "A kinematic error is present.",
    -78: "The motor phase process failed (phase search or phase set from position offset).",
    -79: "The motor phase is not found.",
    -80: "Timeout waiting for control off status.",
    -81: "The requested kinematic mode (number) is not defined for the machine.",
    -82: "Timeout waiting for phase found status.",
    -1000: "Internal error: 'RET_Dev_CfS_NaNReturned'.",
    -1001: "Internal error: 'RET_Dev_CfS_FctNotAvailableInKernel'.",
    -1002: "Internal error: 'RET_Dev_CfS_UndefinedCfSType'.",
    -1003: "Internal error: 'RET_Dev_CfS_FIO_UndefinedFioType'.",
    -1004: "Internal error: 'RET_Dev_CfS_FIO_HomeFile_UndefinedAction'.",
    -1005: "Internal error: 'RET_Dev_UndefinedEnumValue'.",
    -1006: "Internal error: 'RET_Dev_LdataCmdStatusIsNegative'.",
    -1007: "Internal error: 'RET_Dev_NumMotorsInCoord_Sup_DEF_aGrQ_SIZE'.",
    -1008: "Internal error: 'RET_Dev_NumMotorsInCoord_WrongNumber'.",
    -1009: "Internal error: 'RET_String_StrCat_DestSizeReached'.",
    -1010: "Internal error: 'RET_String_LengthOverStringSize'.",
    -1011: "Internal error: 'RET_String_AllCharShouldIntBetween_0_255'.",
    -1012: "Internal error: 'RET_String_StrCpy_DestSizeReached'.",
    -1013: "Internal error: 'RET_ErrAction_HomeReset'.",
    -1014: "Internal error: 'RET_Home_StopReceivedWhileRunning'.",
    -1015: "Internal error: 'RET_UndefinedKinAssembly'.",
    -1016: "Internal error: 'RET_WrongPmcConfig'.",
}


VALIDATION_LIMITS = [
    "Factory workspace limits",
    "Machine workspace limits",
    "User workspace limits",
    "Actuator limits",
    "Joints limits",
    "Due to backlash compensation",
]


def process_cmd_string(command: str) -> str:
    """
    Prepares the command string for sending to the controller.
    A carriage return and newline is appended to the command.
    """

    return add_cr_lf(command)


def wait_until_cmd_is_zero(transport: DeviceTransport, timeout: float = 1.0, interval: float = 0.01):
    """
    Waits until the `cmd` register is 0 (zero) and returns successfully if it does.
    When the `cmd` register doesn't turn zero within the given timeout, a Failure is returned.
    """
    rc = 0

    def c_cmd():
        nonlocal rc
        rc_s = transport.query('c_cmd\r\n').decode()
        LOGGER.debug(f"{rc_s = } <- c_cmd in get_pars")
        if not rc_s.startswith('c_cmd'):
            LOGGER.warning(f"{rc_s = }")
            return -1
        rc = int(rc_s.split('\r\n')[1])
        return rc

    if wait_until(lambda: c_cmd() == 0, interval=interval, timeout=timeout):
        try:
            LOGGER.warning(f"Command check timed out: {RETURN_CODES[rc]} [{rc=}]")
            return Failure(f"Command check resulted in the following error: {RETURN_CODES[rc]}")
        except KeyError:
            return Failure(f"Command check resulted in an unknown error code: {rc = }.")

    return Success("Command finished successfully")


def get_pars(
        transport: DeviceTransport = None, response: bytes = None,
        index: int = 0, count: int = 1, increment: int = 1, timeout: float = 1.0, interval: float = 0.01
) -> bytes | Failure:
    """
    Retrieve the response from a given command from the `c_par` array. The `c_par` array will
    contain the correct values only after the `cmd` register is 0 (zero). So, we will wait until
    `cmd` becomes 0, then retrieve the requested parameters from the `c_par` array and return
    them as a normal response to be processed.

    This function is intended to be used as a `post_cmd` function in the dynamic command decorator.

    Args:
        transport: the transport interface to communicate with the device
        response: the response from the actual command that was sent
        index: starting index for the `c_par` array
        count: the number of c_par values to retrieve
        increment: the increment in the c_par array
        timeout: the timeout period while waiting for the `cmd` to become 0
        interval: the sampling interval for the `cmd`

    Returns:
        The requested values from the `c_par` array as a string. This string can be decoded with
        the `decode_pars()` function.
    """

    if transport is None:
        raise RuntimeError("no device transport was passed into the function!")

    rc = wait_until_cmd_is_zero(transport=transport, interval=interval, timeout=timeout)
    if isinstance(rc, Failure):
        return rc

    # The string 'c_par(0),1,1' is considered illegal.

    if count == 1 and increment == 1:
        query = f"c_par({index})"
    else:
        query = f"c_par({index}),{count},{increment}"

    response = transport.query(f'{query}\r\n')

    LOGGER.debug(f"{response = } <- {query} in get_pars")

    return response


def return_command_status(transport: DeviceTransport = None, response: bytes = None,
                          timeout: float = 1.0, interval: float = 0.01) -> Tuple[int, str] | Failure:
    """
    Check the status of last sent command. When the status is zero (0) the command was successfully executed.
    When the status is not zero before timeout seconds, the status check will be aborted and the function
    will then return the latest status code from the `c_cmd` variable.

    Note: This function shall be used as a `post_cmd` callable in the dynamic_command decorator and the process_response
          shall not be used in conjunction with this command.

    Args:
        transport: the device transport that can be used send additional commands to the device
        response: The response from the command that was sent to the device (ignored by this function)
        timeout: number of seconds before a timeout will occur
        interval: sleep time between checks for the status condition

    Returns:
        The status of the last sent command as a tuple (status code, description).

    Raises:
        A RuntimeError is raised when no device transport is passed into this function.
    """

    # The response argument is ignored in this function as it generates a new response from the
    # command return code.

    if transport is None:
        raise RuntimeError("no device transport was passed into the function!")

    rc = 0

    def c_cmd():
        nonlocal rc
        rc_s = transport.query('c_cmd\r\n').decode()
        LOGGER.debug(f"{rc_s = } <- c_cmd in check_command_status")
        if not rc_s.startswith('c_cmd'):
            LOGGER.warning(f"{rc_s = }")
            return -1
        rc = int(rc_s.split('\r\n')[1])
        return rc

    if wait_until(lambda: c_cmd() == 0, interval=interval, timeout=timeout):
        LOGGER.warning(f"Command check: {RETURN_CODES[rc]} [{rc}]")
    else:
        LOGGER.debug("Success!")
        rc = 0

    return rc, RETURN_CODES[rc]


def check_command_status(transport: DeviceTransport = None, response: bytes = None,
                         timeout: float = 1.0, interval: float = 0.01) -> Any:
    """
    Check the state of last sent command. When the state is zero (0) the command was successfully executed. When the
    status is not zero before timeout seconds, the status check will be aborted and a warning message will be issued.
    This function will then return a Failure containing the latest status code from the `c_cmd` variable.

    The status can take the following values:

    * > 1: a new command has been written but has not yet been interpreted by the communication application
    * = 1: the last written command is currently under execution
    * = 0: the last command was successfully executed
    * < 0: the last command failed. This status code can be used to determine the cause of the error from
      the RETURN_CODES variable in this module.

    Note: This function shall be used as a `post_cmd` callable in the dynamic_command decorator.

    Args:
        transport: the device transport that can be used send additional commands to the device
        response: The response from the command that was sent to the device
        timeout: number of seconds before a timeout will occur
        interval: sleep time between checks for the status condition

    Returns:
        This function passes through the response without change. When the status check timed out
        a Failure will be returned with an associated error message.

    Raises:
        A RuntimeError is raised when no device transport is passed into this function.
    """
    if transport is None:
        raise RuntimeError("no device transport was passed into the function!")

    rc = wait_until_cmd_is_zero(transport=transport, interval=interval, timeout=timeout)
    return rc if isinstance(rc, Failure) else response


def decode_response(response: bytes) -> str | Failure:
    """Decodes the bytes object, strips off the trailing 'CRLF'."""

    LOGGER.debug(f"{response = } <- decode_response")

    return response.decode().rstrip()


def validate_response(response: str, cmd: str = None) -> List[str] | Failure:
    """
    Performs a number of checks on the response string and returns the response as a list of strings.
    The command string (which is the first part of the device response) is removed from the returned list.

    Args:
        response: decoded response from the device
        cmd: the command string that is returned by the device

    Returns:
        A list of strings containing the split response without the first item (which was the command string).
        If the response contains the string 'error #', a Failure will be returned with the error message.
        If the response doesn't start with the given command string (when cmd != None), a Failure is returned.
    """
    if "error #" in response:
        msg = response.split("\r\n")[-1]  # this will strip off the cmd part of the response
        return Failure(msg)

    if cmd is not None and not response.startswith(cmd):
        return Failure(f"Unexpected response from '{cmd}' command: {response}")

    LOGGER.debug(f"{response = } <- validate_response")

    return response.split("\r\n")[1:]


def process_response(response: bytes, cmd: str = None) -> List[str] | Failure:
    """This function is a shortcut for decode_response() and validate_response()."""

    # You might think this shortcut is really useless and doesn't give us anything,
    # the thing is that this function is used in the decorator, which is not possible
    # for each function individually.

    return validate_response(decode_response(response), cmd)


def process_validate_ptp(response: bytes) -> Tuple[int, Dict[int, str]] | Failure:
    """
    The response is in this case the value of 'c_par(0)' which is returned by the `post_cmd` function `get_pars`.

    Args:
        response:

    Returns:

    """
    response = process_response(response)

    if isinstance(response, Failure):
        return response

    response = int(response[0])

    if response == 0:
        return 0, {}

    if response > 0:
        description = decode_validation_error(response)
        return response, description

    # When response is negative, the validation failed, and we extract the command return code

    msg = f"{RETURN_CODES.get(response, 'unknown error code')}"

    LOGGER.error(f"Validate position: error code={response} - {msg}")

    return response, {response: msg}


def issue_warning(*, response, msg: str):
    LOGGER.warning(f"{msg} {response}")
    return response


from typing import TypeVar

T = TypeVar('T', float, int)      # Declare type variable


def decode_pars(response: bytes = None, index: int = 0, count: int = 1, func: Callable = float) -> List[T] | Failure:
    response = process_response(response)

    if isinstance(response, Failure):
        return response

    return [func(x) for x in response[index:index+count]]


def decode_info(response: bytes) -> str | Failure:
    response = process_response(response)

    if isinstance(response, Failure):
        return response

    LOGGER.debug(f"{response = } <- decode_info")

    return (
        f"Info about the Hexapod Alpha+ Controller:\n"
        f"  Software version = {response[1]}.{response[2]}.{response[3]}.{response[4]}\n"
        f"  API version = {response[5]}.{response[6]}.{response[7]}.{response[8]}\n"
        f"  System Configuration version = {response[11]}"
    )


def decode_version(response: bytes) -> str | Failure:
    response = validate_response(decode_response(response))

    if isinstance(response, Failure):
        return response

    return f"{response[5]}.{response[6]}.{response[7]}.{response[8]}"


def decode_uto(response: bytes) -> List[float] | Failure:
    response = validate_response(decode_response(response), "s_uto")

    if isinstance(response, Failure):
        return response

    return [float(x) for x in response]


def decode_mtp(response: bytes) -> List[float] | Failure:
    response = validate_response(decode_response(response), "s_mtp")

    if isinstance(response, Failure):
        return response

    return [float(x) for x in response]


def decode_general_state(response: bytes) -> Tuple[Dict, List] | Failure:
    response = validate_response(decode_response(response), "s_hexa")

    if isinstance(response, Failure):
        return response

    response = int(response[0])

    LOGGER.debug(f"{response = } <- decode_general_state")

    s_hexa = [int(x) for x in f'{response:015b}'[::-1]]
    state = dict(zip(GENERAL_STATE, s_hexa))

    return state, list(state.values())


def decode_actuator_state(response: bytes) -> Tuple[Tuple[Dict, List]] | Failure:
    response = validate_response(decode_response(response), "s_ax")

    if isinstance(response, Failure):
        return response

    def decode_state(state: int) -> Tuple[Dict, List]:
        state_bits = [int(x) for x in f'{state:015b}'[::-1]]
        state_dict = dict(zip(ACTUATOR_STATE, state_bits))
        return state_dict, state_bits

    actuator_states = [int(x) for x in response]

    return tuple(decode_state(state) for state in actuator_states)


def decode_validation_error(value) -> Dict:
    """
    Decode the bitfield variable that is returned by the VALID_PTP command.

    Each bit in this variable represents a particular error in the validation of a movement.
    Several errors can be combined into the given variable.

    Returns a dictionary with the bit numbers that were (on) and the corresponding error description.
    """

    return {bit: VALIDATION_LIMITS[bit] for bit in range(6) if value >> bit & 0b01}


class AlphaPlusTelnetInterface(DeviceTransport, DeviceConnectionInterface):
    """
    The Hexapod controller device interface based on the telnet protocol. This class implements the
    DeviceTransport protocol which provides the `read()`, `write()`, `trans()`, and `query()` methods.
    """

    TELNET_TIMEOUT = 1.0

    def __init__(self, hostname: str = 'localhost', port: int = 23):
        """
        Args:
            hostname (str): the IP address or fully qualified hostname of the OGSE hardware
                controller. The default is 'localhost'.
            port (int): the IP port number to connect to. The default is 23.
        """
        super().__init__()
        self.telnet = Telnet()
        self._is_connected = False
        self.hostname = hostname
        self.port = port

    def connect(self) -> None:
        """
        Connects to the Alpha+ Controller using the Telnet protocol. After connection
        the telnet session logs in with the username provided in the Settings file. The password for this login is also
        provided in the Settings under the same group. Make sure their values are only given in
        the local settings file, not the global settings.

        After login, the `gpascii` command is started with the option `-2` as instructed
        in the software manual of the controller. The first command sent then is the
        `echo7` command, which configures the system to return variable numbers only,
        not their variable names.
        """
        try:
            self.telnet.open(self.hostname, self.port)
        except ConnectionRefusedError as exc:
            raise DeviceConnectionError(
                device_name="Alpha+ Controller",
                message=f"Connection refused to {self.hostname} port {self.port}"
            ) from exc

        try:
            rc = self.telnet.read_until(b"login: ", timeout=self.TELNET_TIMEOUT)
            # print(rc.decode(), flush=True, end="")
            self.telnet.write(f"{PUNA_PLUS.user_name}\r\n".encode())
            rc = self.telnet.read_until(b"Password: ", timeout=self.TELNET_TIMEOUT)
            # print(rc.decode(), flush=True, end="")
            self.telnet.write(f"{PUNA_PLUS.password}\r\n".encode())
            rc = self.telnet.read_until(b"ppmac# ", timeout=self.TELNET_TIMEOUT)
            # print(rc.decode(), flush=True, end="")
            self.telnet.write(b"gpascii -2\r\n")
            rc = self.telnet.read_until(b'\x06\r\n', timeout=self.TELNET_TIMEOUT)
            # print(rc.decode(), flush=True, end="")
            self.telnet.write(b"echo7\r\n")
            rc = self.telnet.read_until(b'\x06\r\n', timeout=self.TELNET_TIMEOUT)
            # print(rc.decode(), flush=True, end="")
        except EOFError as exc:
            raise DeviceConnectionError(
                device_name="Alpha+ Controller",
                message=f"Telnet connection closed for {self.hostname} port {self.port}"
            ) from exc

        self._is_connected = True

    def is_connected(self):
        return self._is_connected

    def disconnect(self):
        rc = self.telnet.read_very_eager()
        print(rc.decode(), flush=True, end="")
        self.telnet.close()
        self._is_connected = False

    def reconnect(self):

        if self._is_connected:
            self.disconnect()
        self.connect()

    def trans(self, cmd: str) -> bytes:
        """
        Send a command to the Aplha+ Controller and waits for a response.
        The response is returned after the ACK is stripped off (see `read()` method).

        Args:
            cmd: a valid command string for the Alpha+ Controller

        Returns:
            The response from the controller on the command that was sent.
        """
        self.write(cmd)
        response = self.read()

        LOGGER.debug(f"trans: {response = }")

        return response

    def read(self) -> bytes:
        """
        Reads a response from the controller.

        Note: The acknowledgement `\x06\r\n` is stripped from the response before it is
              returned. If no ACK is present, a warning message will be logged.

        Returns:
            The response from the controller on a previously sent command.
        """

        response = self.telnet.read_until(b'\x06\r\n', timeout=self.TELNET_TIMEOUT)

        LOGGER.debug(f"read: {response = }")

        if not response.endswith(b'\x06\r\n'):
            LOGGER.warning(f"Expected ACK at the end of the response, {response = }")
            return response

        return response[:-3]  # strip off the ACK

    def write(self, cmd: str):
        """
        Sends a command string to the Alpha+ Controller.
        The command string shall not end with a CRLF, that is automatically appended
        by this function.

        Args:
            cmd: a valid command string for the Alpha+ Controller

        Returns:
            Nothing is returned.
        """
        LOGGER.debug(f"Executing: {cmd.rstrip()}")
        self.telnet.write(cmd.encode())


class AlphaControllerInterface(DeviceInterface):

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cmd=C_STOP",
        process_cmd_string=process_cmd_string,
        process_response=partial(issue_warning, msg="STOP command has been executed."),
        post_cmd=return_command_status,
    )
    def stop(self) -> Tuple[int, str] | Failure:
        """
        Stop the current motion. This command can be sent during a motion of the Hexapod
        and is executed with high priority.

        Returns:
            A tuple (return code, description). Return code = 0 on success.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cmd=C_CONTROLON",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def activate_control_loop(self):
        """
        Activates the control loop on motors.

        It activates the power on the motors and releases the brakes if present.
        The hexapod status 'Control On' will switch to true when the command is successful.

        This command should be used before starting a movement.

        Note: it is possible to activate the control loop on motors even if the home is not complete.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cmd=C_CONTROLOFF",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def deactivate_control_loop(self):
        """
        Disables the control loop on the servo motors.

        It is advisable to disable the servo motors if the system is not used for
        a long period (more than 1 hour for example). However, this recommendation
        depends on the application for which the system is being used.

        This command is performed only if the following conditions are met:
          * there is no motion task running
          * there is no action running
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cmd=C_HOME",
        process_cmd_string=process_cmd_string,
        process_response=process_response,
    )
    def homing(self):
        """
        Starts the homing task on the hexapod.

        Homing needs to be completed before performing any movement. When the hexapod
        is equipped with absolute encoders, this cycle is executed automatically during
        the controller initialization. When the hexapod is not equipped with absolute
        encoders, the homing request movements: homing cycle will search actuators
        reference sensors.

        Homing is required before performing a control movement. Without absolute encoders,
        the homing is performed with a hexapod movement until detecting the reference sensor
        on each of the actuators. The Hexapod will go to a position were the sensors are
        reached that signal a known calibrated position and then returns to the zero position.

        Whenever a homing is performed, the method will return before the actual movement
        is finished.

        The homing cycle takes about two minutes to complete, but the ``homing()`` method
        returns almost immediately. Therefore, to check if the homing is finished, use
        the is_homing_done() method.

        This command is performed only if the following conditions are met:
          * there is no motion task running
          * the emergency stop button is not engaged (not applicable for absolute encoders)

        """
        raise NotImplementedError

    def is_homing_done(self) -> bool | Failure:
        """
        Checks if Homing is done.

        When this variable indicates 'Homing is done' it means the command has been properly
        executed, but it doesn't mean the Hexapod is in position. The hexapod might still be
        moving to its zero position.

        Returns:
            True when the homing is done, False otherwise.
        """
        general_state = self.get_general_state()
        if isinstance(general_state, Failure):
            return general_state

        return bool(general_state[1][HOME_COMPLETE])

    def is_in_position(self) -> bool | Failure:
        """
        Checks if the hexapod is in position.

        Returns:
            True when in position, False otherwise.
        """
        general_state = self.get_general_state()
        if isinstance(general_state, Failure):
            return general_state

        return bool(general_state[1][IN_POSITION]) and not bool(general_state[1][IN_MOTION])

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cmd=C_CLEARERROR",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def clear_error(self) -> Tuple[int, str] | Failure:
        """
        Clear all errors in the controller software.

        This command clears the error list on the controller and automatically removes the error bit of the hexapod
        state. After this command, errors might automatically be regenerated if they are still present. For example,
        if an encoder is disconnected, the encoder error will be re-generated after the command because error reason
        is not corrected.

        Returns:
            The command status is returned as a tuple with (return code, message).
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cfg=1 "
                                           "c_par(0)=${tx_u} c_par(1)=${ty_u} c_par(2)=${tz_u} "
                                           "c_par(3)=${rx_u} c_par(4)=${ry_u} c_par(5)=${rz_u} "
                                           "c_par(6)=${tx_o} c_par(7)=${ty_o} c_par(8)=${tz_o} "
                                           "c_par(9)=${rx_o} c_par(10)=${ry_o} c_par(11)=${rz_o} "
                                           "c_cmd=C_CFG_CS",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def configure_coordinates_systems(
            self, tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o
    ) -> Tuple[int, str] | Failure:
        """
        Change the definition of the User Coordinate System and the Object Coordinate System.

        The parameters tx_u, ty_u, tz_u, rx_u, ry_u, rz_u are used to define the user coordinate
        system relative to the Machine Coordinate System and the parameters tx_o, ty_o, tz_o, rx_o,
        ry_o, rz_o are used to define the Object Coordinate System relative to the Platform
        Coordinate System.

        Args:
            tx_u (float): translation parameter that define the user coordinate system relative
                          to the machine coordinate system [in mm]
            ty_u (float): translation parameter that define the user coordinate system relative
                          to the machine coordinate system [in mm]
            tz_u (float): translation parameter that define the user coordinate system relative
                          to the machine coordinate system [in mm]

            rx_u (float): rotation parameter that define the user coordinate system relative to
                          the machine coordinate system [in deg]
            ry_u (float): rotation parameter that define the user coordinate system relative to
                          the machine coordinate system [in deg]
            rz_u (float): rotation parameter that define the user coordinate system relative to
                          the machine coordinate system [in deg]

            tx_o (float): translation parameter that define the object coordinate system relative
                          to the platform coordinate system [in mm]
            ty_o (float): translation parameter that define the object coordinate system relative
                          to the platform coordinate system [in mm]
            tz_o (float): translation parameter that define the object coordinate system relative
                          to the platform coordinate system [in mm]

            rx_o (float): rotation parameter that define the object coordinate system relative to
                          the platform coordinate system [in deg]
            ry_o (float): rotation parameter that define the object coordinate system relative to
                          the platform coordinate system [in deg]
            rz_o (float): rotation parameter that define the object coordinate system relative to
                          the platform coordinate system [in deg]

        Returns:
            A tuple with the command status return code and a description.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="c_cfg=0 c_cmd=C_CFG_CS",
        process_cmd_string=process_cmd_string,
        process_response=partial(decode_pars, count=12),
        post_cmd=partial(get_pars, count=12),
    )
    def get_coordinates_systems(self):
        """
        Retrieve the definition of the User Coordinate System and the Object Coordinate System.

        Returns:
            tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o where the
            translation parameters are in [mm] and the rotation parameters are in [deg].
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_par(0)=${cm} "
                                           "c_par(1)=${tx} c_par(2)=${ty} c_par(3)=${tz} "
                                           "c_par(4)=${rx} c_par(5)=${ry} c_par(6)=${rz} "
                                           "c_cmd=C_MOVE_PTP",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def move_ptp(self, cm: int,
                 tx: float, ty: float, tz: float, rx: float, ry: float, rz: float) -> Tuple[int, str] | Failure:
        """
        Start the movement as defined by the arguments.

        Args:
            cm: control mode, 0=absolute, 1=object relative, 2=user relative
            tx: position on X-axis [mm]
            ty: position on Y-axis [mm]
            tz: position on Z-axis [mm]
            rx: rotation around the X-axis [deg]
            ry: rotation around the Y-axis [deg]
            rz: rotation around the Z-axis [deg]

        Returns:

        """
        raise NotImplementedError

    def move_absolute(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, str] | Failure:
        return self.move_ptp(0, tx, ty, tz, rx, ry, rz)

    def move_relative_object(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, str] | Failure:
        return self.move_ptp(1, tx, ty, tz, rx, ry, rz)

    def move_relative_user(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, str] | Failure:
        return self.move_ptp(2, tx, ty, tz, rx, ry, rz)

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_par(0)=${pos} c_cmd=C_MOVE_SPECIFICPOS",
        process_cmd_string=process_cmd_string,
        # process_response=process_response,
        post_cmd=return_command_status,
    )
    def goto_specific_position(self, pos: int):
        raise NotImplementedError

    def goto_user_zero_position(self):
        return self.goto_specific_position(pos=1)

    def goto_retracted_position(self):
        return self.goto_specific_position(pos=2)

    def goto_machine_zero_position(self):
        return self.goto_specific_position(pos=3)

    goto_zero_position = goto_user_zero_position

    @dynamic_command(
        cmd_type="transaction", cmd_string="s_uto_tx,6,1",
        process_cmd_string=process_cmd_string,
        process_response=decode_uto,
    )
    def get_user_positions(self):
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="s_mtp_tx,6,1",
        process_cmd_string=process_cmd_string,
        process_response=decode_mtp,
    )
    def get_machine_positions(self) -> List[float] | Failure:
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="c_cmd=C_VERSION",
        process_cmd_string=process_cmd_string,
        process_response=decode_info,
        post_cmd=partial(get_pars, count=12),
    )
    def info(self) -> str:
        """Returns basic information about the hexapod and the controller.

        Returns:
            a multiline response message containing the device info.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="c_cmd=C_VERSION",
        process_cmd_string=process_cmd_string,
        process_response=decode_version,
        post_cmd=partial(get_pars, count=12),
    )
    def version(self) -> str:
        """Returns the version of the firmware running on the hexapod aplha+ controller.

        Returns:
            A version number as a string.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="s_hexa",
        process_cmd_string=process_cmd_string,
        process_response=decode_general_state
    )
    def get_general_state(self) -> Tuple[Dict, List] | Failure:
        """
        Asks the general state of the hexapod on all the motors following the bits definition
        presented below.

        GENERAL_STATE =
            0: "Error",
            1: "System initialized",
            2: "Control on",
            3: "In position",
            4: "Motion task running",
            5: "Home task running",
            6: "Home complete",
            7: "Home virtual",
            8: "Phase found",
            9: "Brake on",
            10:"Motion restricted",
            11:"Power on encoders",
            12:"Power on limit switches",
            13:"Power on drives",
            14:"Emergency stop"

        Returns:
            A dictionary with the bits value of each parameter.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="s_ax_1,6,1",
        process_cmd_string=process_cmd_string,
        process_response=decode_actuator_state,
    )
    def get_actuator_state(self) -> Tuple[Dict, List] | Failure:
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="s_pos_ax_1,6,1",
        process_cmd_string=process_cmd_string,
        process_response=partial(decode_pars, count=6, func=float),
    )
    def get_actuator_length(self):
        """
        Retrieve the current length of the hexapod actuators.

        Returns:
            array: an array of six float values for actuator length L1 to L6 in [mm], and \
            None: when an Exception was raised and logs the error message.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="c_cfg=0 c_cmd=C_CFG_SPEED",
        process_cmd_string=process_cmd_string,
        process_response=partial(decode_pars, count=6),
        post_cmd=partial(get_pars, count=6),
    )
    def get_speed(self) -> List[float]:
        """
        Returns the positional speed of movements.

        Returns a list of floating point numbers [vt, vr, vt-, vr-, vt+, vr+] where vt and vr are the translation and
        angular speed respectively, the '-' and '+' are the minimum and maximum speeds.
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cfg=1 c_par(0)=${vt} c_par(1)=${vr} c_cmd=C_CFG_SPEED",
        process_cmd_string=process_cmd_string,
        post_cmd=return_command_status,
    )
    def set_speed(self, vt: float, vr: float):
        """
        Set the positioning speed of movements.

        Args:
            vt: translational speed, unit = mm per second
            vr: angular speed, unit = deg per second
        """
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_par(0)=${vm} c_par(1)=${cm} "
                                           "c_par(2)=${tx} c_par(3)=${ty} c_par(4)=${tz} "
                                           "c_par(5)=${rx} c_par(6)=${ry} c_par(7)=${rz} "
                                           "c_cmd=C_VALID_PTP",
        process_cmd_string=process_cmd_string,
        process_response=process_validate_ptp,
        post_cmd=get_pars,
    )
    def validate_position(self, vm, cm, tx, ty, tz, rx, ry, rz) -> Tuple[int, Dict[int, str]] | Failure:
        """
        Ask the controller if the movement defined by the arguments is feasible.

        Returns a tuple where the first element is an integer that represents the
        bitfield encoding the errors. The second element is a dictionary with the
        bit numbers that were (on) and the corresponding error description as
        defined by VALIDATION_LIMITS.

        Args:
            vm (int): validation mode [only vm=1 is currently implemented by Symétrie]
            cm (int): control mode (0 = absolute, 1 = object relative, 2 = user relative)
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        """
        raise NotImplementedError

    def check_absolute_movement(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, Dict[int, str]] | Failure:
        """
        Check if the requested object movement is valid.

        The absolute movement is expressed in the user coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.
        """
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 0 for absolute
        return self.validate_position(1, 0, tx, ty, tz, rx, ry, rz)

    def check_relative_object_movement(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, Dict[int, str]] | Failure:
        """
        Check if the requested object movement is valid.

        The relative motion is expressed in the object coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.
        """
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 1 for object relative
        return self.validate_position(1, 1, tx, ty, tz, rx, ry, rz)

    def check_relative_user_movement(self, tx, ty, tz, rx, ry, rz) -> Tuple[int, Dict[int, str]] | Failure:
        """
        Check if the requested object movement is valid.

        The relative motion is expressed in the user coordinate system.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        Returns:
            tuple: where the first element is an integer that represents the
                bitfield encoding the errors. The second element is a dictionary
                with the bit numbers that were (on) and the corresponding error
                description.
        """
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 2 for user relative
        return self.validate_position(1, 2, tx, ty, tz, rx, ry, rz)


class AlphaPlusControllerInterface(AlphaControllerInterface):
    @dynamic_command(
        cmd_type="query", cmd_string="${name}",
        process_cmd_string=process_cmd_string,
        process_response=process_response,
    )
    def query_variable(self, name: str) -> str:
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="${name},${count},${increment}",
        process_cmd_string=process_cmd_string,
        process_response=process_response,
    )
    def query_variables(self, name: str, count: int, increment: int = 1):
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="${name}(${idx})",
        process_cmd_string=process_cmd_string,
        process_response=process_response,
    )
    def query_array(self, name: str, idx: int):
        raise NotImplementedError

    @dynamic_command(
        cmd_type="query", cmd_string="${name}(${idx}),${count},${increment}",
        process_cmd_string=process_cmd_string,
        process_response=process_response,
    )
    def query_array_values(self, name: str, idx: int, count: int, increment: int = 1):
        raise NotImplementedError

    @dynamic_command(
        cmd_type="transaction", cmd_string="c_cfg=0 c_par(0)=${lim} c_cmd=C_CFG_LIMIT",
        process_cmd_string=process_cmd_string,
        process_response=partial(decode_pars, index=1, count=12),
        post_cmd=partial(get_pars, count=13),
    )
    def get_limits_value(self, lim):
        """
        Three different and independent operational workspace limits are defined on the controller:

            * Factory limits: are expressed in machine coordinate system limits. Those parameters cannot be modified.
            * Machine coordinate system limits: they are expressed in the Machine coordinate system. It can be used to
              secure the hexapod (and/or object) from its environment.
            * User coordinate system limits: they are expressed in the User coordinate system. It can be used to limits
              the movements of the object mounted on hexapod.

        Remark: operational workspace limits must be understood as limits in terms of amplitude of movement. Those
        limits are defined for each operational axis with a negative and positive value and are used in the validation
        process. Position on each operational axis must be within those two values.

        Args:
            lim (int): 0 = factory (GET only), 1 = machine cs limit, 2 = user cs limit

        Returns:
            A list of 12 float values: tx-, ty-, tz-, rx-, ry-, rz-, tx+, ty+, tz+, rx+, ry+, rz+
            The first six values are the negative limits for translation and rotation, the last six numbers are the
            positive limits for translation and rotation.

        """
        raise NotImplementedError


if __name__ == "__main__":

    from rich import print as rp

    from egse.hexapod.symetrie.punaplus import PunaPlusController

    # Thie IP address for the emulator running in VirtualBox is 192.168.56.10
    # When you run the emulator in Parallels, the IP address is 10.37.129.10

    puna = PunaPlusController(hostname="10.37.129.10", port=23)
    puna.connect()

    with Timer("PunaPlusController"):
        rp(puna.info())
        rp(puna.version())
        rp(puna.is_homing_done())
        rp(puna.homing())
        if wait_until(puna.is_homing_done, interval=1, timeout=300):
            rp("[red]Task puna.is_homing_done() timed out after 30s.[/red]")
        rp(puna.is_homing_done())
        rp(puna.is_in_position())
        rp(puna.activate_control_loop())
        rp(puna.get_general_state())
        rp(puna.get_actuator_state())
        rp(puna.deactivate_control_loop())
        rp(puna.get_general_state())
        rp(puna.get_actuator_state())
        rp(puna.stop())
        rp(puna.get_limits_value(0))
        rp(puna.get_limits_value(1))
        rp(puna.check_absolute_movement(1, 1, 1, 1, 1, 1))
        rp(puna.check_absolute_movement(51, 51, 51, 1, 1, 1))
        rp(puna.get_speed())
        rp(puna.set_speed(2.0, 1.0))
        rp(speed := puna.get_speed())

        if speed[:2] != [2.0, 1.0]:
            rp(f"[red]Expected {speed[:2]} == [2.0, 1.0][/red")

        input("Check speed parameters in GUI")

        rp(puna.set_speed(1.2, 1.1))

        rp(speed := puna.get_speed())

        if speed[:2] != [1.2, 1.1]:
            rp(f"[red]Expected {speed[:2]} == [1.2, 1.1][/red")

        rp(puna.get_actuator_length())

        # rp(puna.machine_limit_enable(0))
        # rp(puna.machine_limit_enable(1))
        # rp(puna.get_limits_state())
        rp(puna.get_coordinates_systems())
        rp(puna.configure_coordinates_systems(
                0.033000, -0.238000, 230.205000, 0.003282, 0.005671, 0.013930,
                0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000))
        rp(puna.get_coordinates_systems())
        rp(puna.get_machine_positions())
        rp(puna.get_user_positions())

        input("Check configuration in GUI")

        rp(puna.configure_coordinates_systems(
                0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
                0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000))

        rp(puna.validate_position(1, 0, 0, 0, 0, 0, 0, 0))
        rp(puna.validate_position(1, 0, 0, 0, 50, 0, 0, 0))

        rp(puna.goto_zero_position())
        rp(puna.is_in_position())
        if wait_until(puna.is_in_position, interval=1, timeout=300):
            rp("[red]Task puna.is_in_position() timed out after 30s.[/red]")
        rp(puna.is_in_position())

        rp(puna.get_machine_positions())
        rp(puna.get_user_positions())

        rp(puna.move_absolute(0, 0, 12, 0, 0, 10))

        rp(puna.is_in_position())
        if wait_until(puna.is_in_position, interval=1, timeout=300):
            rp("[red]Task puna.is_in_position() timed out after 30s.[/red]")
        rp(puna.is_in_position())

        rp(puna.get_machine_positions())
        rp(puna.get_user_positions())

        rp(puna.move_absolute(0, 0, 0, 0, 0, 0))

        rp(puna.is_in_position())
        if wait_until(puna.is_in_position, interval=1, timeout=300):
            rp("[red]Task puna.is_in_position() timed out after 30s.[/red]")
        rp(puna.is_in_position())

        rp(puna.get_machine_positions())
        rp(puna.get_user_positions())

        # # puna.reset()
        puna.disconnect()

        # rp(0, decode_validation_error(0))
        # rp(11, decode_validation_error(11))
        # rp(8, decode_validation_error(8))
        # rp(24, decode_validation_error(24))



from egse.setup import load_setup


def test_private_caching_functions(default_env):

    from egse.confman import _get_cached_setup_info
    from egse.confman import _populate_cached_setup_info
    from egse.confman import _print_cached_setup_info
    from egse.confman import _add_setup_info_to_cache

    assert _get_cached_setup_info(0) is None

    _populate_cached_setup_info()

    assert _get_cached_setup_info(0) is not None
    assert _get_cached_setup_info(0).site_id == "LAB23"

    _print_cached_setup_info()

    from egse.env import print_env
    print_env()

    # The Setup shall be loaded from the configuration data location, not from the repo location (which is None)
    setup = load_setup(setup_id=28, site_id="LAB23", from_disk=True)

    assert setup.get_id() == '00028'
    assert 'SETUP_LAB23_00028' in str(setup.get_filename())

    _add_setup_info_to_cache(setup)

    assert _get_cached_setup_info(28) is not None
    assert _get_cached_setup_info(28).site_id == 'LAB23'
    assert _get_cached_setup_info(28).path.name.endswith('SETUP_LAB23_00028_240123_120028.yaml')

    _print_cached_setup_info()



"""
This module implements a random walk that can be used in simulators for devices like temperature sensors or power meter.
"""
from __future__ import annotations

import random
import sys
from typing import Optional


# Re-implement based on:
#
# 1. https://isquared.digital/blog/2020-04-12-random-walk/
# 2. https://isquared.digital/blog/2020-04-16-brownian-motion/
# 3. https://isquared.digital/blog/2020-05-01-drifted-brownian-motion/

class RandomWalk:
    """
    This class implements a random walk.

    The RandomWalk is an iterator which can be used in a for loop:

    ```Python
    import matplotlib.pyplot as plt
    values = [value for value in RandomWalk(start=15.0, boundary=(-100, 25), scale=0.01, count=10000)]
    plt.plot(values)
    plt.show()
    ```

    The code above will create a plot like below, where 10000 values are plot in a random walk between -100 and 25,
    starting at 15:

    ![Random Walk example](../../img/screenshot-randomwalk.png)

    When you need a random walk that will return values infinitely, set count = 0.

    By default, the returned values will represent a random walk between -1.0 and 1.0 in a step of 1.0.
    """
    def __init__(self, start: float = 0.0, boundary: tuple = (-1.0, 1.0), threshold: tuple | float = 0.5,
                 scale: float = 1.0, count: int = 1000, seed: Optional[int] = None):
        """

        Args:
            start: the start value for the random walk [default = 0.0]
            boundary: the lower and upper boundaries [default = (-1.0, 1.0)]
            threshold: the probability to move up or down [default = 0.5]
            scale: scale the step size [default = 1.0]
            count: the number of iterations, when <= 0 count will be basically infinite [default = 1000]
            seed: seed for the initialization of the random generator [default = None]
        """
        self._threshold_up = threshold[1] if isinstance(threshold, tuple) else threshold
        self._threshold_down = threshold[0] if isinstance(threshold, tuple) else threshold
        self._boundary = boundary
        self._min_value, self._max_value = boundary
        self._count = count if count > 0 else sys.maxsize
        self._start = start
        self._last = start
        self._scale = scale
        self._seed = seed
        self._random = random.Random()

        if seed is not None:
            self._random.seed(seed)

    def __iter__(self):
        return self

    def __next__(self):
        if self._count > 0:
            probability = self._random.random()
            if probability > self._threshold_up:
                distance_to_limit = self._max_value - self._last
                step = self._random.uniform(0.0, distance_to_limit / 10)
                value = self._last + self._scale * step
                value = min(value, self._max_value)
            elif probability < self._threshold_down:
                distance_to_limit = self._last - self._min_value
                step = self._random.uniform(0.0, distance_to_limit / 10)
                value = self._last - self._scale * step
                value = max(value, self._min_value)
            else:
                value = self._last
            self._last = value
            self._count -= 1
            return value
        raise StopIteration

    def x__next__(self):
        if self._count > 0:
            probability = self._random.random()
            if probability > self._threshold_up:
                value = self._last + self._scale
                value = min(value, self._max_value)
            elif probability < self._threshold_down:
                value = self._last - self._scale
                value = max(value, self._min_value)
            else:
                value = self._last
            self._last = value
            self._count -= 1
            return value
        raise StopIteration


if __name__ == "__main__":

    # https://towardsdatascience.com/generating-synthetic-time-series-data-with-random-walks-8701bb9a56a8

    import matplotlib.pyplot as plt

    # Random Walk, each time a different seed by default

    y1 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.50, scale=0.5, count=1_000_000)
    # y1 = RandomWalk(start=50.0, boundary=(0, 100), threshold=(0.25, 0.70), scale=1, count=10_000)
    # y2 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.49, scale=0.05, count=100000, seed=0)
    # y3 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.48, scale=0.01, count=100000, seed=0)
    # plt.plot(list(zip(y1, y2, y3)))
    plt.plot(list(y1))
    plt.show()

    y1 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.5, scale=0.2, count=100000, seed=0)
    y2 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.51, scale=0.2, count=100000, seed=0)
    y3 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.52, scale=0.2, count=100000, seed=0)
    plt.plot(list(zip(y1, y2, y3)))
    plt.show()

    y1 = RandomWalk(start=75.0, boundary=(0, 100), threshold=0.5, scale=0.2, count=100000, seed=0)
    y2 = RandomWalk(start=50.0, boundary=(0, 100), threshold=0.5, scale=0.2, count=100000, seed=0)
    y3 = RandomWalk(start=25.0, boundary=(0, 100), threshold=0.5, scale=0.2, count=100000, seed=0)
    plt.plot(list(zip(y1, y2, y3)))
    plt.show()

    rng0 = RandomWalk(start=15.0, boundary=(-20, 25), scale=1.0, count=1000, seed=1)
    rng1 = RandomWalk(start=15.0, boundary=(-20, 25), scale=0.8, count=1000, seed=10)
    rng2 = RandomWalk(start=15.0, boundary=(-20, 25), scale=0.5, count=1000, seed=100)
    rng3 = RandomWalk(start=15.0, boundary=(-20, 25), scale=0.3, count=1000, seed=1000)

    values = list(zip(rng0, rng1, rng2, rng3))
    plt.plot(values)
    plt.ylim(-20, 25)
    plt.show()

    # Example 1:

    rw = RandomWalk(start=25.0, boundary=(-100, 100), scale=0.2, count=0, seed=42)
    values = [next(rw) for _ in range(1000000)]
    plt.plot(values)
    plt.show()

    # Example 2: should give the same output as example 1

    plt.plot(
        list(RandomWalk(start=25.0, boundary=(-100, 100), scale=0.2, count=1000000, seed=42,))
    )
    plt.show()

    # Example 3:

    for i, value in enumerate(RandomWalk(start=25.0, boundary=(-100, 100), threshold=0.5, scale=2, count=20)):
        print(f"{i}, {value}")



import logging
import time
from pathlib import Path

import pytest
from egse.decorators import borg
from egse.decorators import classproperty
from egse.decorators import debug
from egse.decorators import deprecate
from egse.decorators import dynamic_interface
from egse.decorators import profile
from egse.decorators import profile_func
from egse.decorators import query_command
from egse.decorators import read_command
from egse.decorators import singleton
from egse.decorators import spy_on_attr_change
from egse.decorators import static_vars
from egse.decorators import time_it
from egse.decorators import timer
from egse.decorators import to_be_implemented
from egse.decorators import transaction_command
from egse.decorators import write_command
from egse.settings import Settings

MODULE_LOGGER = logging.getLogger(__name__)


def test_static_vars():

    @static_vars(counter=-1)
    def increment(amount: int = 1):
        increment.counter += amount
        return increment.counter

    assert increment() == 0
    assert increment(5) == 5
    assert increment() == 6


def test_timer(caplog):

    caplog.set_level(logging.INFO)

    @timer(level=logging.INFO, precision=1)
    def func(seconds: float = 1.0):
        time.sleep(seconds)

    func()
    assert "Finished 'func' in 1.0 secs" in caplog.text

    func(2.3)
    assert "Finished 'func' in 2.3 secs" in caplog.text


def test_time_it(caplog):

    caplog.set_level(logging.INFO)

    @time_it(count=1000_000)
    def func(x: float = 1.0):
        return x ** 2

    func()
    assert "Finished 'func' in " in caplog.text


def test_debug(caplog):
    caplog.set_level(level=logging.DEBUG)

    @debug
    def func1():
        pass

    func1()
    assert "Calling func1()" in caplog.text

    @debug
    def func2(x, y):
        return x*y

    func2(2, 3)
    assert "Calling func2(2, 3)" in caplog.text

    func2('#', 5)
    assert "Calling func2('#', 5)" in caplog.text
    assert "'func2' returned '#####'"


def test_to_be_implemented(caplog):

    @to_be_implemented
    def func():
        pass

    func()
    assert 'WARNING' in caplog.text
    assert 'func is not yet implemented'


def test_profile(capsys):

    @profile
    def func(name: str = "World"):
        return f"hello, {name}!"

    Settings.set_profiling(False)

    func()

    captured = capsys.readouterr()
    assert "PROFILE" not in captured.out

    Settings.set_profiling(True)

    func()

    captured = capsys.readouterr()
    assert "PROFILE" in captured.out

    Settings.set_profiling(False)


def test_profile_func(caplog):

    caplog.set_level(level=logging.DEBUG)

    def power(x: float, exp: int = 2):
        return x ** exp

    output_file = "profiling_func.prof"

    @profile_func(output_file=output_file, strip_dirs=True, sort_by=('calls', 'cumulative'))
    def func(x: int = 0):
        return power(x)

    func()

    assert Path(output_file).exists()

    # TODO: add some more checks on the content of the profiling output

    Path(output_file).unlink()


def test_dynamic_interface():

    @dynamic_interface
    def func():
        return func.__dynamic_interface
    assert func() is True

    @query_command
    def query():
        return query.__query_command
    assert query() is True

    @transaction_command
    def transaction():
        return transaction.__transaction_command
    assert transaction() is True

    @read_command
    def read():
        return read.__read_command
    assert read() is True

    @write_command
    def write():
        return write.__write_command
    assert write() is True


def test_classproperty():
    class Message:
        _msg_cache = set()

        def __init__(self):
            self.prefix = "msg> "

        @classproperty
        def messages_count(cls):
            return len(cls._msg_cache)

        @classproperty
        def name(cls):
            return cls.__name__

        @classmethod
        def add_message(cls, msg):
            cls._msg_cache.add(msg)

    msg = Message()

    assert msg.messages_count == 0

    msg.add_message("message 1")

    assert msg.messages_count == 1

    assert "Message" in msg.name

    with pytest.raises(AttributeError):
        msg.name = "SomethingElse"

    assert "Something" not in msg.name


def test_borg():
    @borg
    class BorgOne:
        _dry_run = False

        @classmethod
        def is_dry_run(cls) -> bool:
            return cls._dry_run

        @classmethod
        def set_dry_run(cls, flag: bool):
            cls._dry_run = flag

    assert BorgOne.is_dry_run() is False
    assert BorgOne().is_dry_run() is False

    BorgOne.set_dry_run(True)

    assert BorgOne.is_dry_run() is True
    assert BorgOne().is_dry_run() is True

    BorgOne().set_dry_run(False)
    assert BorgOne.is_dry_run() is False
    assert BorgOne().is_dry_run() is False

    assert BorgOne() is not BorgOne()  # it's not a singleton


def test_singleton():
    @singleton
    class Foo:
        def __new__(cls):
            cls.x = 10
            return object.__new__(cls)

        def __init__(self):
            assert self.x == 10
            self.x = 15

    assert Foo().x == 15

    foo = Foo()

    assert foo.x == 15

    foo.x = 20

    assert foo.x == 20
    assert Foo().x == 20

    assert Foo() == foo == Foo()
    assert Foo() is foo is Foo()


def test_deprecation():
    """
    Simple tests for the @deprecate decorator.

    If you want to check how the warning message looks like, you can use

        with pytest.deprecated_call() as record:
            deprecated_function()
            MODULE_LOGGER.debug(record[0].message)

    """

    @deprecate(
        reason="we are testing the @deprecate decorator " "on a simple function without arguments",
        alternative="the new_function() instead",
    )
    def deprecated_function():
        # delegate to new_function()
        return new_function()

    def new_function():
        pass

    # This is how we check that indeed the Warning was issued?

    with pytest.deprecated_call():
        deprecated_function()

    class Foo:
        @deprecate(
            reason="we need to test if the @deprecate decorator also works "
            "properly with (class) methods"
        )
        def is_deprecated(self):
            return True

    foo = Foo()

    with pytest.deprecated_call():
        foo.is_deprecated()


def test_spy_on_attr_change(caplog):

    class X:
        ...

    class Y:
        ...

    # Check if the changes are reported in the log messages

    spy_on_attr_change(x := X())

    x.a = 42
    assert "in X -> a: <Nothing> -> 42" in caplog.text
    x.b = 37
    assert "in X -> b: <Nothing> -> 37" in caplog.text
    x.a = 5
    assert "in X -> a: 42 -> 5" in caplog.text

    caplog.clear()

    xx = X()

    # Another instance of class X should not be monitored

    xx.aa = 77
    assert "in xx -> aa: <Nothing> -> 77" not in caplog.text

    spy_on_attr_change(xx, obj_name="xx")

    # now this instance is also monitored and change reported in the log

    xx.aa = 55
    assert "in xx -> aa: 77 -> 55" in caplog.text

    caplog.clear()

    spy_on_attr_change(y := Y())

    # These tests are mainly to confirm that the name of the class is correct
    # even with multiple calls to the `spy_on_attr_change()` function.

    x.b = 3
    assert "in X -> b: 37 -> 3" in caplog.text
    y.a = 8
    assert "in Y -> a: <Nothing> -> 8" in caplog.text
    xx.bb = 11
    assert "Spy: in xx -> bb: <Nothing> -> 11" in caplog.text



# An example plugin for the `cgse {start,stop,status} service` command from `cgse-core`.
#
import subprocess
import sys
import time
from pathlib import Path
from typing import Annotated

import rich
import typer

daq6510 = typer.Typer(
    name="daq6510",
    help="DAQ6510 Data Acquisition Unit, Keithley, temperature monitoring",
    no_args_is_help=True
)


@daq6510.command(name="start")
def start_daq6510():
    """Start the daq6510 service."""
    rich.print("Starting service daq6510")


@daq6510.command(name="stop")
def stop_daq6510():
    """Stop the daq6510 service."""
    rich.print("Terminating service daq6510")


@daq6510.command(name="status")
def status_daq6510():
    """Print status information on the daq6510 service."""

    proc = subprocess.Popen(
        [sys.executable, '-m', 'egse.tempcontrol.keithley.daq6510_sim', 'status'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.DEVNULL
    )

    stdout, stderr = proc.communicate()

    rich.print(stdout.decode(), end='')
    # if stderr:
    #     rich.print(f"[red]{stderr.decode()}[/]")


@daq6510.command(name="start-sim")
def start_daq6510_sim():
    """Start the DAQ6510 Simulator."""
    rich.print("Starting service DAQ6510 Simulator")

    out = open(Path('~/.daq6510_sim.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.tempcontrol.keithley.daq6510_sim', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


@daq6510.command(name="stop-sim")
def stop_daq6510_sim():
    """Stop the DAQ6510 Simulator."""
    rich.print("Terminating the DAQ6510 simulator.")

    out = open(Path('~/.daq6510_sim.stop.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.tempcontrol.keithley.daq6510_sim', 'stop'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


if __name__ == '__main__':
    daq6510()



"""
This module implements a standard Observer <-> Observable pattern.

!!! Bug
    The methods in the Observable are not PEP8 compliant and need to be deprecated and changed.
"""
import abc


class Observer(abc.ABC):
    """The observer that needs to take action when notified."""
    @abc.abstractmethod
    def update(self, changed_object):
        pass

    @abc.abstractmethod
    def do(self, actions):
        pass


class Observable:
    """The object that sends out notifications to the observers. """
    def __init__(self):
        self.observers = []

    def addObserver(self, observer):
        if observer not in self.observers:
            self.observers.append(observer)

    def deleteObserver(self, observer):
        self.observers.remove(observer)

    def clearObservers(self):
        self.observers = []

    def countObservers(self):
        return len(self.observers)

    def notifyObservers(self, changedObject):
        # FIXME: put a try..except here to log any problem that occurred in the observer's update()
        #        method
        for observer in self.observers:
            observer.update(changedObject)

    def actionObservers(self, actions):
        # FIXME: put a try..except here to log any problem that occurred in the observer's do()
        #        method
        for observer in self.observers:
            observer.do(actions)



import ctypes
import logging

import pytest

from egse.bits import beautify_binary
from egse.bits import bit_set
from egse.bits import bits_set
from egse.bits import clear_bit
from egse.bits import clear_bits
from egse.bits import crc_calc
from egse.bits import humanize_bytes
from egse.bits import s16
from egse.bits import s32
from egse.bits import set_bit
from egse.bits import set_bits
from egse.bits import toggle_bit

logger = logging.getLogger(__name__)


def test_clear_bit():
    bf = 0b11111111

    for bit in range(8):
        assert bit_set(bf, bit)

    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")

    bf = clear_bit(bf, 3)
    assert not bit_set(bf, 3)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")

    bf = clear_bit(bf, 0)
    assert not bit_set(bf, 0)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")

    bf = clear_bit(bf, 7)
    assert not bit_set(bf, 7)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")


def test_set_bit():
    bf = 0b00000000

    for bit in range(8):
        assert not bit_set(bf, bit)

    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")
    bf = set_bit(bf, 3)
    assert bit_set(bf, 3)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")
    bf = set_bit(bf, 0)
    assert bit_set(bf, 0)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")
    bf = set_bit(bf, 7)
    assert bit_set(bf, 7)
    logger.debug(f"bf = {bf:4}, 0b{bf:08b}")
    assert bits_set(bf, 0, 3, 7)
    assert bits_set(bf, [0, 3, 7])
    assert not bits_set(bf, [0, 1, 3, 7])

    assert bits_set(bf, 0)
    assert bits_set(bf, [0])


def test_toggle_bit():

    bf = 0b0000_0000

    assert toggle_bit(bf, 0) == 0b0001
    assert toggle_bit(bf, 7) == 0b1000_0000

    bf = toggle_bit(bf, 3)

    assert bf == 0b0000_1000

    bf = 0b0000_0000

    for i in range(8):
        bf = toggle_bit(bf, i)

    assert bf == 0b1111_1111


def test_beautify_binary():
    assert beautify_binary(0b0010, group=4) == '0000 0010'
    assert beautify_binary(0b1111_1111, group=4) == '1111 1111'
    assert beautify_binary(0b0000_0001_1111_1111) == '00000001 11111111'

    assert beautify_binary(2 ** 7) == '10000000'
    assert beautify_binary(2 ** 8) == '00000001 00000000'
    assert beautify_binary(2 ** 9 - 1, group=4, sep='_') == '0000_0001_1111_1111'
    assert beautify_binary(2 ** 14) == '01000000 00000000'
    assert beautify_binary(2 ** 24) == '00000001 00000000 00000000 00000000'
    assert beautify_binary(2 ** 27) == '00001000 00000000 00000000 00000000'
    assert beautify_binary(2 ** 33) == '00000010 00000000 00000000 00000000 00000000'

    assert beautify_binary(0b11111111000011110000111100000000, sep='_') == \
        '11111111_00001111_00001111_00000000'
    assert beautify_binary(0b11111111000011110000111100000000, sep='_', group=4) == \
        '1111_1111_0000_1111_0000_1111_0000_0000'

    assert beautify_binary(0b0101_0101_0101_0011, sep='_', group=4, prefix='0b') == \
        '0b0101_0101_0101_0011'


def test_set_bits():
    assert set_bits(0, (0, 1)) == 0b00000001
    assert set_bits(0, (7, 8)) == 0b10000000
    assert set_bits(0, (1, 3)) == 0b00000110
    assert set_bits(0, (0, 8)) == 0b11111111

    assert set_bits(0b0001_1000, (3, 5)) == 0b0001_1000
    assert set_bits(0b0001_1000, (2, 6)) == 0b0011_1100

    assert set_bits(0, (128, 129)) == 2 ** 128
    assert set_bits(0, (0, 129)) == 2 ** 129 - 1


def test_alternative_set_bits():

    for width in 1, 4, 8, 12, 15, 16, 32, 128, 129:
        assert set_bits(0, (0, width)) == int('1'*width, 2)


def test_clear_bits():
    assert clear_bits(0xFF, (0, 1)) == 0b11111110
    assert clear_bits(0xFF, (7, 8)) == 0b01111111
    assert clear_bits(0xFF, (1, 3)) == 0b11111001
    assert clear_bits(0xFF, (0, 8)) == 0b00000000

    assert clear_bits(0b0001_1000, (3, 5)) == 0b0000_0000
    assert clear_bits(0b0011_1100, (3, 5)) == 0b0010_0100


def test_crc_calc():
    b = b'abcdefghijklmnopqrstuvwxyz'

    assert crc_calc(b, 0,  0) == 0x00
    assert crc_calc(b, 0,  1) == 0xD9
    assert crc_calc(b, 0, 16) == 0x8A
    assert crc_calc(b, 0, 26) == 0x63

    assert crc_calc(b, 5,  0) == 0x00
    assert crc_calc(b, 5,  1) == 0xAC
    assert crc_calc(b, 5, 16) == 0x12

    with pytest.raises(IndexError):
        assert crc_calc(b, 5, 26) == 0x32

    with pytest.raises(IndexError):
        assert crc_calc(b, 0, 27) == 0x63

    b = ctypes.create_string_buffer(b'abcdefghijklmnopqrstuvwxyz')

    assert crc_calc(b, 0,  0) == 0x00
    assert crc_calc(b, 0,  1) == 0xD9
    assert crc_calc(b, 0, 16) == 0x8A
    assert crc_calc(b, 0, 26) == 0x63

    assert crc_calc(b, 5,  0) == 0x00
    assert crc_calc(b, 5,  1) == 0xAC
    assert crc_calc(b, 5, 16) == 0x12

    with pytest.raises(IndexError):
        assert crc_calc(b, 5, 26) == 0x32

    # Note that we have an additional \x0 (NUL) character here added by the
    # ctypes function because we passed in a bytes object. So, the IndexError
    # will only be generated when accessing 28 characters!

    with pytest.raises(IndexError):
        assert crc_calc(b, 0, 28) == 0x63

def test_humanize_bytes():

    assert humanize_bytes(0) == "0 bytes"
    assert humanize_bytes(1) == "1 byte"
    assert humanize_bytes(60) == "60 bytes"
    assert humanize_bytes(1023) == "1023 bytes"
    assert humanize_bytes(1024) == "1.000 KiB"
    assert humanize_bytes(1025) == "1.001 KiB"
    assert humanize_bytes(1024*2 - 1) == "1.999 KiB"
    assert humanize_bytes(1024*2) == "2.000 KiB"
    assert humanize_bytes(1024*2 + 1) == "2.001 KiB"

    assert humanize_bytes(1024**2 - 1) == "1023.999 KiB"
    assert humanize_bytes(1024**2) == "1.000 MiB"
    assert humanize_bytes(1024**2 + 1) == "1.000 MiB"
    assert humanize_bytes(1024**2 + 100) == "1.000 MiB"
    assert humanize_bytes(1024**2 + 1000) == "1.001 MiB"

    assert humanize_bytes(1024**3 - 1000) == "1023.999 MiB"
    assert humanize_bytes(1024**3) == "1.000 GiB"
    assert humanize_bytes(1024**3 + 1) == "1.000 GiB"
    assert humanize_bytes(1024**3 + 100) == "1.000 GiB"
    assert humanize_bytes(1024**3 + 1000) == "1.000 GiB"
    assert humanize_bytes(1024**3 + 1000000) == "1.001 GiB"

    assert humanize_bytes(1024**8) == "1.000 YiB"
    assert humanize_bytes(1024**9) == "1237940039285380274899124224 bytes"

    assert humanize_bytes(0, base=10) == "0 bytes"
    assert humanize_bytes(1, base=10) == "1 byte"
    assert humanize_bytes(60, base=10) == "60 bytes"
    assert humanize_bytes(999, base=10) == "999 bytes"
    assert humanize_bytes(1000, base=10) == "1.000 kB"
    assert humanize_bytes(1024, base=10) == "1.024 kB"
    assert humanize_bytes(1000*2 - 1, base=10) == "1.999 kB"
    assert humanize_bytes(1000*2, base=10) == "2.000 kB"
    assert humanize_bytes(1000*2 + 1, base=10) == "2.001 kB"

    assert humanize_bytes(10**6 - 1, base=10) == "999.999 kB"
    assert humanize_bytes(10**6, base=10) == "1.000 MB"
    assert humanize_bytes(10**6 + 1, base=10) == "1.000 MB"
    assert humanize_bytes(10**6 + 100, base=10) == "1.000 MB"
    assert humanize_bytes(10**6 + 1000, base=10) == "1.001 MB"
    assert humanize_bytes(10**6 + 1, base=10, precision=6) == "1.000001 MB"
    assert humanize_bytes(10**6 + 100, base=10, precision=6) == "1.000100 MB"
    assert humanize_bytes(10**6 + 1000, base=10, precision=6) == "1.001000 MB"

    assert humanize_bytes(10**9 - 1000, base=10) == "999.999 MB"
    assert humanize_bytes(10**9, base=10) == "1.000 GB"
    assert humanize_bytes(10**9 + 1, base=10) == "1.000 GB"
    assert humanize_bytes(10**9 + 100, base=10) == "1.000 GB"
    assert humanize_bytes(10**9 + 1000, base=10) == "1.000 GB"
    assert humanize_bytes(10**9 + 1000000, base=10) == "1.001 GB"
    assert humanize_bytes(10**9 + 1000000, base=10, precision=1) == "1.0 GB"

    assert humanize_bytes(10**12, base=10) == "1.000 TB"
    assert humanize_bytes(10**15, base=10) == "1.000 PB"
    assert humanize_bytes(10**18, base=10) == "1.000 EB"
    assert humanize_bytes(10**21, base=10) == "1.000 ZB"
    assert humanize_bytes(10**24, base=10) == "1.000 YB"

    assert humanize_bytes(1024**9, base=10) == "1237940039285380274899124224 bytes"

    # other bases than 2 or 10 are not supported

    assert 1073741824 == 8**10 == 1024**3

    with pytest.raises(ValueError):
        assert humanize_bytes(8**10, base=8) == "1.000 GiB"

    assert humanize_bytes(1024**3, base=2) == "1.000 GiB"
    assert humanize_bytes(1073741824, base=10) == "1.074 GB"

    assert humanize_bytes(1024**3, base='binary') == "1.000 GiB"
    assert humanize_bytes(1073741824, base='decimal') == "1.074 GB"


def test_s16():

    assert s16(0b1000_0000_0001_0001) == -32751
    assert s16(0b1000_0000_0000_0001) == -32767
    assert s16(0b1111_1111_1111_1111) == -1
    assert s16(0b0111_1111_1111_1111) == 32767


def test_s32():

    assert s32(0b1000_0000_0000_0000_0000_0000_0001_0001) == -2147483631
    assert s32(0b1000_0000_0000_0000_0000_0000_0000_0001) == -2147483647
    assert s32(0b1111_1111_1111_1111_1111_1111_1111_1111) == -1
    assert s32(0b0111_1111_1111_1111_1111_1111_1111_1111) == 2147483647



"""
This script bumps the version of all libs and projects in this monorepo to the version that
is currently in the `pyproject.toml` file in the root folder of the monorepo.

Usage:
    $ python bump.py <part>

where `<part>` should be 'patch', 'minor', or 'major'.

Note:
    You are expected to be in the minimal virtual environment associated with this monorepo,
    being a `pyenv` or Poetry environment, or your global environment shall include the tomlkit
    and the rich package.

    Alternatively (and preferably) you can use `uv run build.py <part>` to run this script.

"""

# /// script
# requires-python = ">=3.9"
# dependencies = [
#   "tomlkit",
#   "rich",
#   "typer",
# ]
# ///
import os
import pathlib
import sys

import rich
import tomlkit
import tomlkit.exceptions
import typer


def bump_version(version, part='patch'):
    major, minor, patch = map(int, version.split('.'))

    if part == 'major':
        major += 1
        minor = 0
        patch = 0
    elif part == 'minor':
        minor += 1
        patch = 0
    elif part == 'patch':
        patch += 1
    else:
        raise ValueError("Part must be 'major', 'minor', or 'patch'")

    return f"{major}.{minor}.{patch}"


def get_master_version(master_pyproject_path):
    """Returns the version number of the master project, i.e. cgse."""

    with open(master_pyproject_path, "r") as file:
        data = tomlkit.parse(file.read())

    return data["project"]["version"]


def update_project_version(project_dir, new_version):
    """Updates the version of the subproject."""

    os.chdir(project_dir)

    # Check if the Poetry version is defined, otherwise print a message.

    with open("pyproject.toml", "r") as file:
        data = tomlkit.parse(file.read())

    try:
        data["project"]["version"] = new_version

        with open("pyproject.toml", "w") as file:
            tomlkit.dump(data, file)

    except tomlkit.exceptions.NonExistentKey:
        rich.print(f"[red]\[project.version] is not defined in pyproject.toml in {project_dir}[/]")


def update_all_projects_in_monorepo(root_dir: pathlib.Path, part: str, dry_run: bool = False, verbose: bool = True):
    """
    Updates all pyproject.toml files with the master version number.

    Parameters:
        root_dir:
        part:
        dry_run:
        verbose:
    """

    excluded_subdirs = ["__pycache__", ".venv", ".git", ".idea", "cgse/build", "cgse/dist"]

    master_version = get_master_version(os.path.join(root_dir, "pyproject.toml"))

    new_version = bump_version(master_version, part=part)

    rich.print(f"Projects will be bumped from version {master_version} to version {new_version}")

    for subdir, dirs, files in os.walk(root_dir):
        if subdir == "." or subdir == ".." or any(excluded in subdir for excluded in excluded_subdirs):
            # rich.print(f"rejected {subdir = }")
            continue
        if subdir != str(root_dir):
            # continue  # skip the root project file
            pass
        if "pyproject.toml" in files:
            verbose and print(f"Updating version for project in {subdir}")
            if not dry_run:
                update_project_version(subdir, new_version)


app = typer.Typer()


@app.command()
def main(part: str, dry_run: bool = False, verbose: bool = True):

    monorepo_root = pathlib.Path(__file__).parent.resolve()

    cwd = os.getcwd()
    os.chdir(monorepo_root)

    try:
        update_all_projects_in_monorepo(monorepo_root, part, dry_run, verbose)
    except ValueError as exc:
        rich.print(f"[red]{exc.__class__.__name__}: {exc}[/]")
        sys.exit(1)
    finally:
        os.chdir(cwd)


if __name__ == "__main__":
    app()




class CGSEException(Exception):
    """The base exception for all errors and warnings in the Common-EGSE."""

    pass


class Error(CGSEException):
    """The base class for all Common-EGSE Errors."""

    pass


class Warning(CGSEException):
    """The base class for all Common-EGSE Warnings."""

    pass


class FileIsEmptyError(Error):
    """Raised when a file is empty and that is unexpected."""
    pass


class InvalidOperationError(Error):
    """
    Raised when a certain operation is not valid in the given state,
    circumstances or environment.
    """

    pass


class InvalidInputError(Error):
    """ Exception raised when the input is invalid after editing."""

    pass


class DeviceNotFoundError(Error):
    """Raised when a device could not be located, or loaded."""

    pass


class InternalStateError(Error):
    """Raised when an object encounters an internal state inconsistency."""

    pass


class InternalError(Error):
    """Raised when an internal inconsistency occurred in a function, method or class."""

    pass


class Abort(RuntimeError):
    """Internal Exception to signal a process to abort."""



"""
This module defines the device classes to be used to connect to and control the Hexapod ZONDA from
Symétrie.

"""
import logging
import math
import time
from typing import Dict

from egse.device import DeviceInterface
from egse.hexapod import HexapodError
from egse.hexapod.symetrie.alpha import AlphaPlusControllerInterface
from egse.hexapod.symetrie.zonda_devif import RETURN_CODES
from egse.hexapod.symetrie.zonda_devif import ZondaError
from egse.hexapod.symetrie.zonda_devif import ZondaTelnetInterface
from egse.hexapod.symetrie.zonda_devif import decode_command
from egse.proxy import Proxy
from egse.settings import Settings
from egse.system import Timer
from egse.system import wait_until
from egse.zmq_ser import connect_address
from numpy import loadtxt
from time import sleep

logger = logging.getLogger(__name__)

ZONDA_SETTINGS = Settings.load("ZONDA Controller")
CTRL_SETTINGS = Settings.load("Hexapod ZONDA Control Server")
DEVICE_SETTINGS = Settings.load(filename="zonda.yaml")

HOME_COMPLETE = 6
IN_POSITION = 3
IN_MOTION = 4

GENERAL_STATE = [
    "Error",
    "System initialized",
    "Control on",
    "In position",
    "Motion task running",
    "Home task running",
    "Home complete",
    "Home virtual",
    "Phase found",
    "Brake on",
    "Motion restricted",
    "Power on encoders",
    "Power on limit switches",
    "Power on drives",
    "Emergency stop",
]

ACTUATOR_STATE = [
    "Error",
    "Control on",
    "In position",
    "Motion task running",
    "Home task running",
    "Home complete",
    "Phase found",
    "Brake on",
    "Home hardware input",
    "Negative hardware limit switch",
    "Positive hardware limit switch",
    "Software limit reached",
    "Following error",
    "Drive fault",
    "Encoder error",
]

VALIDATION_LIMITS = [
    "Factory workspace limits",
    "Machine workspace limits",
    "User workspace limits",
    "Actuator limits",
    "Joints limits",
    "Due to backlash compensation",
]


class ZondaInterface(AlphaPlusControllerInterface, DeviceInterface):
    """
    Interface definition for the ZondaController, the ZondaProxy and the ZondaSimulator.
    """


class ZondaController(ZondaInterface):
    def __init__(self):

        super().__init__()

        logger.info(f"Initializing ZondaController with hostname={ZONDA_SETTINGS.IP} on port={ZONDA_SETTINGS.PORT}")

        self.v_pos = None

        self.cm = ["absolute", "object relative", "user relative"]

        self.speed = {"vt": 0,
                      "vr": 0,
                      "vtmin": 0,
                      "vrmin": 0,
                      "vtmax": 0,
                      "vrmax": 0}

        try:
            self.hexapod = ZondaTelnetInterface()

        except ZondaError as exc:
            logger.warning(
                f"HexapodError: Couldn't establish connection with the Hexapod ZONDA Hardware "
                f"Controller: ({exc})"
            )

    def is_simulator(self):
        return False

    def is_connected(self):
        return self.hexapod.is_connected()

    def connect(self):
        try:
            self.hexapod.connect(ZONDA_SETTINGS.IP)
        except ZondaError as exc:
            logger.warning(f"ZondaError caught: Couldn't establish connection ({exc})")
            raise HexapodError("Couldn't establish a connection with the Hexapod.") from exc

    def disconnect(self):
        try:
            self.hexapod.disconnect()
        except ZondaError as exc:
            raise HexapodError("Couldn't disconnect from Hexapod.") from exc

    def reconnect(self):
        if self.is_connected():
            self.disconnect()
        self.connect()

    def info(self):
        try:
            _version = self.hexapod.trans("VERS")
            msg = "Info about the Hexapod ZONDA:\n"
            msg += f"version = {_version}\n"
        except ZondaError as exc:
            raise HexapodError(
                "Couldn't retrieve information from Hexapod ZONDA Hardware Controller."
            ) from exc

        return msg

    def get_general_state(self):
        """
        Asks the general state of the hexapod on all the motors following the bits definition
        presented below.

        GENERAL_STATE =
            0: "Error",
            1: "System initialized",
            2: "Control on",
            3: "In position",
            4: "Motion task running",
            5: "Home task running",
            6: "Home complete",
            7: "Home virtual",
            8: "Phase found",
            9: "Brake on",
            10:"Motion restricted",
            11:"Power on encoders",
            12:"Power on limit switches",
            13:"Power on drives",
            14:"Emergency stop"

        Returns:
            A dictionary with the bits value of each parameter.
        """
        try:
            rc = self.hexapod.trans("s_hexa")
            rc = int(rc[0])

            # from int to bit list of 15 elements corresponding to the hexapod state bits
            # the bit list must be reversed to get lsb

            s_hexa = [int(x) for x in f'{rc:015b}'[::-1]]
            state = {k: v for k, v in zip(GENERAL_STATE, s_hexa)}
        except ZondaError as exc:
            raise HexapodError("Couldn't retrieve the state from Hexapod.") from exc

        return [state, list(state.values())]

    def stop(self):
        try:
            self.hexapod.trans("c_cmd=C_STOP")
            sc = self.hexapod.check_command_status()
            logger.warning(f"Stop command has been executed: {sc}")
        except ZondaError as exc:
            raise HexapodError("Couldn't disconnect from Hexapod.") from exc
        return sc

    def clear_error(self):
        try:
            print("Cleaning errors from buffer")
            sc = self.hexapod.trans("c_cmd=C_CLEARERROR")
            number = self.hexapod.trans('s_err_nr')
            number = int(number[0])

            if number == 0:
                logger.info("All the errors have been cleared.")
            else:
                logger.warning("Couldn't clear all errors from Hexapod.")

        except ZondaError as exc:
            raise HexapodError("Couldn't clear all the errors from Hexapod.") from exc
        return sc

    def reset(self, wait=True):
        try:
            print("Resetting the Hexapod Controller")
            print("STOP and CONTROLOFF commands are sent to the controller before resetting...")
            self.stop()
            print("Hexapod is stopped")
            self.deactivate_control_loop()
            print("Hexapod control loop has been deactivated")

            print("Rebooting the controller: will take 2 min to initialize")

            # FIXME: you might want to rethink this, because the reboot will close the Ethernet
            #     connection and therefore we need to reconnect after a two minutes waiting time.
            #     During the waiting time, the GUI should have all functions disabled and the
            #     connection icon should show a disconnected icon. Don't use `time.sleep()` as
            #     that will block the GUI, run a timer in a QThread signaling a reconnect when
            #     the sleep time is over.
            self.hexapod.trans("system reboot")

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return None

    def jog(self, axis: int, inc: float):
        try:
            self.hexapod.trans("c_ax={} c_par(0)={} c_cmd=C_JOG")
            sc = self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def perform_maintenance(self, axis):
        try:
            if axis == 10:
                self.hexapod.trans("c_par(0)=3 c_cmd=C_MAINTENANCE")
            else:
                self.hexapod.trans("c_par(0)=2 c_par(1)={} c_cmd=C_MAINTENANCE".format(str(axis)))

            sc = self.hexapod.check_command_status()
        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def set_default(self):
        try:
            print("Resetting the hexapod to the factory default parameters...:")
            self.hexapod.trans("c_cfg=1 c_cmd=C_CFG_DEFAULT")
            sc = self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def homing(self):
        try:
            print("Executing homing command")
            self.hexapod.trans("c_cmd=C_HOME")
            sc = self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def goto_specific_position(self, pos):
        try:
            sc = self.hexapod.trans("c_par(0)={} c_cmd=C_MOVE_SPECIFICPOS".format(str(pos)))
            print(sc)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return sc

    def goto_zero_position(self):
        try:
            sc = self.hexapod.trans("c_par(0)=1 c_cmd=C_MOVE_SPECIFICPOS")
            print(sc)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return sc

    def goto_retracted_position(self):
        try:
            sc = self.hexapod.trans("c_par(0)=2 c_cmd=C_MOVE_SPECIFICPOS")
            print(sc)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return sc

    def is_homing_done(self):
        try:
            sc = self.get_general_state()
            homing = sc[1][HOME_COMPLETE]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return homing

    def is_in_position(self):
        try:
            sc = self.get_general_state()
            in_position = sc[1][IN_POSITION]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return in_position

    def activate_control_loop(self):
        try:
            self.hexapod.trans("c_cmd=C_CONTROLON")
            print("Executing activate_control_loop command")
            sc = self.hexapod.check_command_status()
        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def deactivate_control_loop(self):
        try:
            self.hexapod.trans("c_cmd=C_CONTROLOFF")
            print("Executing deactivate_control_loop")
            sc = self.hexapod.check_command_status()
        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return sc

    def machine_limit_enable(self, state: int):
        try:
            command = decode_command("CFG_LIMITENABLE", 1, state)
            print(f"Executing machine_limit_enable command to state {state}. ")
            self.hexapod.trans(command)
            rc = self.hexapod.check_command_status()
        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return rc

    def machine_limit_set(
            self, tx_n, ty_n, tz_n, rx_n, ry_n, rz_n, tx_p, ty_p, tz_p, rx_p, ry_p, rz_p):
        try:
            name = "CFG_LIMIT"
            arguments = [1, tx_n, ty_n, tz_n, rx_n, ry_n, rz_n, tx_p, ty_p, tz_p, rx_p, ry_p, rz_p]

            cmd = decode_command(name, *arguments)

            print("Executing machine_limit_set command {}...: ".format(cmd), end="")
            self.hexapod.trans(cmd)
            self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

    def user_limit_enable(self, state: int):
        try:
            command = decode_command("CFG_LIMITENABLE", 2, state)
            print(f"Executing user_limit_enable command to state {state}...: ")
            self.hexapod.trans(command)
            self.hexapod.check_command_status()
        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return None

    def user_limit_set(
            self, tx_n, ty_n, tz_n, rx_n, ry_n, rz_n, tx_p, ty_p, tz_p, rx_p, ry_p, rz_p):
        try:
            name = "CFG_LIMIT"
            arguments = [2, tx_n, ty_n, tz_n, rx_n, ry_n, rz_n, tx_p, ty_p, tz_p, rx_p, ry_p, rz_p]

            cmd = decode_command(name, *arguments)

            print(f"Executing user_limit_set command {cmd}...: ")
            self.hexapod.trans(cmd)
            rc = self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return rc

    def __move(self, cm, tx, ty, tz, rx, ry, rz):
        """
        Ask the controller to perform the movement defined by the arguments and the command MOVE_PTP.

        For all control modes cm, the rotation centre coincides with the Object
        Coordinates System origin and the movements are controlled with translation
        components at first (Tx, Ty, tZ) and then the rotation components (Rx, Ry, Rz).

        Control mode cm:
            * 0 = absolute control, object coordinate system position and orientation
                    expressed in the invariant user coordinate system
            * 1 = object relative, motion expressed in the Object Coordinate System
            * 2 = user relative, motion expressed in the User Coordinate System

        Args:
            cm (int): control mode
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]
        """

        name = "MOVE_PTP"
        arguments = [cm, tx, ty, tz, rx, ry, rz]

        command = decode_command(name, *arguments)
        print(f"Executing move_ptp command in {self.cm[cm]} mode ... {command}")
        self.hexapod.trans(command)
        return self.hexapod.check_command_status()

    def validate_position(self, vm, cm, tx, ty, tz, rx, ry, rz):
        # Currently only the vm=1 mode is developed by Symétrie
        """
        Ask the controller if the movement defined by the arguments is feasible.

        Returns a tuple where the first element is an integer that represents the
        bitfield encoding the errors. The second element is a dictionary with the
        bit numbers that were (on) and the corresponding error description as
        defined by VALIDATION_LIMITS.

        Args:
            tx (float): position on the X-axis [mm]
            ty (float): position on the Y-axis [mm]
            tz (float): position on the Z-axis [mm]
            rx (float): rotation around the X-axis [deg]
            ry (float): rotation around the Y-axis [deg]
            rz (float): rotation around the Z-axis [deg]

        """
        name = "VALID_PTP"
        arguments = [vm, cm, tx, ty, tz, rx, ry, rz]

        command = decode_command(name, *arguments)

        #print(f"Executing validate_position command: {command}")

        self.hexapod.trans(command)
        rc = self.hexapod.check_command_status()
        pc = self.hexapod.get_pars(0)
        pc = int(pc[0])

        if pc == 0:
            return 0, {}

        if pc > 0:
            d = decode_validation_error(pc)
            return pc, d

        # When coming here the command returned an error

        msg = f"{RETURN_CODES.get(pc, 'unknown error code')}"
        logger.error(f"Validate position: error code={pc} - {msg}")

        return pc, {pc: msg}

    def move_absolute(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(0, tx, ty, tz, rx, ry, rz)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return rc

    def move_relative_object(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(1, tx, ty, tz, rx, ry, rz)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return rc

    def move_relative_user(self, tx, ty, tz, rx, ry, rz):
        try:
            rc = self.__move(2, tx, ty, tz, rx, ry, rz)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return rc

    def check_absolute_movement(self, tx, ty, tz, rx, ry, rz):
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 0
        return self.validate_position(1, 0, tx, ty, tz, rx, ry, rz)

    def check_relative_object_movement(self, tx, ty, tz, rx, ry, rz):
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 1 for object
        return self.validate_position(1, 1, tx, ty, tz, rx, ry, rz)

    def check_relative_user_movement(self, tx, ty, tz, rx, ry, rz):
        # Currently only the vm=1 mode is developed by Symétrie
        # Parameter cm = 2 for user relative
        return self.validate_position(1, 2, tx, ty, tz, rx, ry, rz)

    def get_temperature(self):
        #TODO: to be tested with the real Hexapod (the emulator does not implement this and only returns zeros)
        try:
            temp = self.hexapod.trans("s_ai_1,6,1")
            temp = [float(x) for x in temp]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return temp

    def get_user_positions(self):
        try:
            uto = self.hexapod.trans("s_uto_tx,6,1")
            uto = [float(x) for x in uto]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return uto

    def get_machine_positions(self):
        try:
            mtp = self.hexapod.trans("s_mtp_tx,6,1")
            mtp = [float(x) for x in mtp]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return mtp

    def get_actuator_length(self):
        try:
            pos = self.hexapod.trans("s_pos_ax_1,6,1")
            pos = [float(x) for x in pos]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return pos

    def get_actuator_state(self):

        response = []
        try:
            actuator_states = self.hexapod.trans("s_ax_1,6,1")
            actuator_states = [int(x) for x in actuator_states]

            for idx, state in enumerate(actuator_states):
                state_bits = [int(x) for x in f'{state:015b}'[::-1]]
                state_dict = {k: v for k, v in zip(ACTUATOR_STATE, state_bits)}
                response.append([state_dict, state_bits])

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return response

    def get_coordinates_systems(self):
        try:
            command = decode_command("CFG_CS?")
            print(f"Executing get_coordinate_system command: {command}")
            self.hexapod.trans(command)
            sc = self.hexapod.check_command_status()
            cs = self.hexapod.get_pars(12)
            cs = [float(x) for x in cs]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return cs

    def configure_coordinates_systems(
        self, tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o
    ):
        try:
            name = "CFG_CS"
            arguments = [tx_u, ty_u, tz_u, rx_u, ry_u, rz_u, tx_o, ty_o, tz_o, rx_o, ry_o, rz_o]
            command = decode_command(name, *arguments)
            print(f"Executing configure_coordinates_systems: {command}")
            self.hexapod.trans(command)
            rc = self.hexapod.check_command_status()

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return rc

    def get_limits_value(self, lim):
        # Not implemented in Puna
        # lim: int | 0 = Factory, 1 = machine cs limits, 2 = user cs limits
        try:
            cmd = decode_command("CFG_LIMIT?", lim)
            print(f"Executing get_limits_value command: {cmd=}")
            self.hexapod.trans(cmd)
            self.hexapod.check_command_status()

            pc = self.hexapod.get_pars(13)
            pc = [float(x) for x in pc[1:]]

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return pc

    def get_limits_state(self):
        # not implemented in Puna
        try:
            command = decode_command("CFG_LIMITENABLE?")
            print(f"Executing get_limits_state command: {command=}")
            self.hexapod.trans(command)
            sc = self.hexapod.check_command_status()

            if sc[0] == 0:
                pc = self.hexapod.get_pars(3)
                pc = [int(i) for i in pc]
            else:
                pc = ["nan", "nan", "nan"]

            keys = VALIDATION_LIMITS[:3]
            ls = {k: v for k, v in zip(keys, pc)}

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc

        return ls

    def get_speed(self):
        try:
            command = decode_command("CFG_SPEED?")
            print(f"Executing get_speed command: {command}")
            self.hexapod.trans(command)
            sc = self.hexapod.check_command_status()
            speed = self.hexapod.get_pars(6)
            speed = [float(x) for x in speed]
            keys = list(self.speed.keys())

            s = {k: v for k, v in zip(keys, speed)}

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return s

    def set_speed(self, vt, vr):
        try:
            name = "CFG_SPEED"
            arguments = [vt, vr]
            command = decode_command(name, *arguments)
            # The parameters are automatically limited by the controller between the factory
            # configured min/max speeds
            self.hexapod.trans(command)

        except ZondaError as exc:
            raise HexapodError("Couldn't Execute command on Hexapod.") from exc
        return

    def sequence(self, file_path: str, time_sleep: float):
        file = file_path + ".txt"
        SEQUENCE = loadtxt(file)
        for step in SEQUENCE:
            state = self.check_absolute_movement(step[0], step[1], step[2], step[3], step[4], step[5])
            if state[0] != 0:
                print('Error: Out of bounds! One point is out of the workspace!')
                return
        print("OK! The entire trajectory is reachable in the defined workspace.")
        step_number = 0
        for step in SEQUENCE:
            step_number += 1
            print("\nExecuting step number ", step_number, "over ", len(SEQUENCE), ": ", step)
            self.move_absolute(step[0], step[1], step[2], step[3], step[4], step[5])
            in_pos = self.get_general_state()[1][3]
            while not(in_pos):
                in_pos = self.get_general_state()[1][3]
            print("Step ", step_number, "done.\nWaiting for ", time_sleep, "seconds.")
            sleep(time_sleep)
        print('Sequence "' + file_path + '" done with success!')


class ZondaSimulator(ZondaInterface):
    """
    HexapodSimulator simulates the Symétrie Hexapod ZONDA. The class is heavily based on the
    ReferenceFrames in the `egse.coordinates` package.

    The simulator implements the same methods as the HexapodController class which acts on the
    real hardware controller in either simulation mode or with a real Hexapod ZONDA connected.

    Therefore, the HexapodSimulator can be used instead of the Hexapod class in test harnesses
    and when the hardware is not available.

    This class simulates all the movements and status of the Hexapod.
    """
    def __init__(self):
        # Keep a record if the homing() command has been executed.

        self.homing_done = False
        self.control_loop = False
        self._virtual_homing = False
        self._virtual_homing_position = None

    def is_simulator(self):
        return True

    def connect(self):
        pass

    def reconnect(self):
        pass

    def disconnect(self):
        # TODO:
        #   Should I keep state in this class to check if it has been disconnected?
        #
        # TODO:
        #   What happens when I re-connect to this Simulator? Shall it be in Homing position or
        #   do I have to keep state via a persistency mechanism?
        pass

    def is_connected(self):
        return True

    def clear_error(self):
        return 0

    def homing(self):
        self.goto_zero_position()
        self.homing_done = True
        self._virtual_homing = False
        self._virtual_homing_position = None
        return 0

    def is_homing_done(self):
        return self.homing_done

    def activate_control_loop(self):
        self.control_loop = True
        return self.control_loop

    def deactivate_control_loop(self):
        self.control_loop = False
        return self.control_loop

    pass


class ZondaProxy(Proxy, ZondaInterface):
    """The ZondaProxy class is used to connect to the control server and send commands to the
    Hexapod ZONDA remotely."""

    def __init__(
            self,
            protocol=CTRL_SETTINGS.PROTOCOL,
            hostname=CTRL_SETTINGS.HOSTNAME,
            port=CTRL_SETTINGS.COMMANDING_PORT,
    ):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address) [default is taken from settings
            file]
            port: TCP port on which the control server is listening for commands [default is
            taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port))


def decode_validation_error(value) -> Dict:
    """
    Decode the bitfield variable that is returned by the VALID_PTP command.

    Each bit in this variable represents a particular error in the validation of a movement.
    Several errors can be combined into the given variable.

    Returns a dictionary with the bit numbers that were (on) and the corresponding error description.
    """

    return {bit: VALIDATION_LIMITS[bit] for bit in range(6) if value >> bit & 0b01}


if __name__ == "__main__":

    from rich import print as rp

    zonda = ZondaController()
    zonda.connect()

    with Timer("ZondaController"):

        rp(zonda.info())
        rp(zonda.is_homing_done())
        rp(zonda.is_in_position())
        rp(zonda.activate_control_loop())
        rp(zonda.get_general_state())
        rp(zonda.get_actuator_state())
        rp(zonda.deactivate_control_loop())
        rp(zonda.get_general_state())
        rp(zonda.get_actuator_state())
        rp(zonda.stop())
        rp(zonda.get_limits_value(0))
        rp(zonda.get_limits_value(1))
        rp(zonda.check_absolute_movement(1, 1, 1, 1, 1, 1))
        rp(zonda.check_absolute_movement(51, 51, 51, 1, 1, 1))
        rp(zonda.get_speed())
        rp(zonda.set_speed(2.0, 1.0))
        time.sleep(0.5)  # if we do not sleep, the get_speed() will get the old values
        speed = zonda.get_speed()

        if not math.isclose(speed['vt'], 2.0):
            rp(f"[red]{speed['vt']} != 2.0[/red]")
        if not math.isclose(speed['vr'], 1.0):
            rp(f"[red]{speed['vr']} != 1.0[/red]")

        rp(zonda.get_actuator_length())

        # rp(zonda.machine_limit_enable(0))
        # rp(zonda.machine_limit_enable(1))
        # rp(zonda.get_limits_state())
        rp(zonda.get_coordinates_systems())
        rp(zonda.configure_coordinates_systems(
            0.033000, -0.238000, 230.205000, 0.003282, 0.005671, 0.013930,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000))
        rp(zonda.get_coordinates_systems())
        rp(zonda.get_machine_positions())
        rp(zonda.get_user_positions())
        rp(zonda.configure_coordinates_systems(
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000,
            0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000))
        rp(zonda.validate_position(1, 0, 0, 0, 0, 0, 0, 0))
        rp(zonda.validate_position(1, 0, 0, 0, 50, 0, 0, 0))

        rp(zonda.goto_zero_position())
        rp(zonda.is_in_position())
        if wait_until(zonda.is_in_position, interval=1, timeout=300):
            rp("[red]Task zonda.is_in_position() timed out after 30s.[/red]")
        rp(zonda.is_in_position())

        rp(zonda.get_machine_positions())
        rp(zonda.get_user_positions())

        rp(zonda.move_absolute(0, 0, 12, 0, 0, 10))

        rp(zonda.is_in_position())
        if wait_until(zonda.is_in_position, interval=1, timeout=300):
            rp("[red]Task zonda.is_in_position() timed out after 30s.[/red]")
        rp(zonda.is_in_position())

        rp(zonda.get_machine_positions())
        rp(zonda.get_user_positions())

        rp(zonda.move_absolute(0, 0, 0, 0, 0, 0))

        rp(zonda.is_in_position())
        if wait_until(zonda.is_in_position, interval=1, timeout=300):
            rp("[red]Task zonda.is_in_position() timed out after 30s.[/red]")
        rp(zonda.is_in_position())

        rp(zonda.get_machine_positions())
        rp(zonda.get_user_positions())

        # zonda.reset()
        zonda.disconnect()

        # rp(0, decode_validation_error(0))
        # rp(11, decode_validation_error(11))
        # rp(8, decode_validation_error(8))
        # rp(24, decode_validation_error(24))



"""
A Graphical User Interface for monitoring and commanding the Symétrie ZONDA Hexapod.

Start the GUI from your terminal as follows:

    zonda_ui [--type proxy|direct|simulator]

This GUI is based on the SYM_positioning application from Symétrie. The intent
is to provide operators a user interface which is platform independent, but
familiar.

The application is completely written in Python/Qt5 and can therefore run on any
platform that supports Python and Qt5.

"""
import argparse
import logging
import multiprocessing
from pathlib import Path

import sys
import threading
from typing import List

multiprocessing.current_process().name = "zonda_ui"

import pyqtgraph as pg
from PyQt5.QtCore import QDateTime, QLockFile
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QMessageBox
from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QGroupBox
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidget
from prometheus_client import start_http_server

from egse.gui import show_warning_message
from egse.gui.led import Indic
from egse.gui.states import States
from egse.gui.stripchart import StripChart
from egse.hexapod.symetrie.hexapod_ui import ActuatorStates
from egse.hexapod.symetrie.hexapod_ui import HexapodUIController
from egse.hexapod.symetrie.hexapod_ui import HexapodUIModel
from egse.hexapod.symetrie.hexapod_ui import HexapodUIView
from egse.hexapod.symetrie.zonda import ZondaController
from egse.hexapod.symetrie.zonda import ZondaProxy
from egse.hexapod.symetrie.zonda import ZondaSimulator
from egse.process import ProcessStatus
from egse.resource import get_resource
from egse.settings import Settings
from egse.system import do_every

MODULE_LOGGER = logging.getLogger(__name__)

# Status LEDs define the number of status leds (length of the list), the description and the
# default color when the LED is on.

STATUS_LEDS = [
    ["Error", Indic.RED],  # bit 0
    ["System Initialized", Indic.GREEN],  # bit 1
    ["Control On", Indic.GREEN],  # bit 2
    ["In Position", Indic.GREEN],  # bit 3
    ["Motion Task Running", Indic.GREEN],  # bit 4
    ["Home Task Running", Indic.GREEN],  # bit 5
    ["Home Complete", Indic.GREEN],  # bit 6
    ["Home Virtual", Indic.GREEN],  # bit 7
    ["Phase Found", Indic.GREEN],  # bit 8
    ["Brake on", Indic.GREEN],  # bit 9
    ["Motion Restricted", Indic.RED],  # bit 10
    ["Power on Encoders", Indic.GREEN],  # bit 11
    ["Power on Limit switches", Indic.GREEN],  # bit 12
    ["Power on Drives", Indic.GREEN],  # bit 13
    ["Emergency Stop", Indic.RED],  # bit 14
]

# The index of the Control LED

CONTROL_ONOFF = 2

ACTUATOR_STATE_LABELS = [
    "Error: ",
    "Control On: ",
    "In Position: ",
    "Motion Task Running: ",
    "Home task running: ",
    "Home complete: ",
    "Phase found: ",
    "Brake on: ",
    "Home HW input: ",
    "Negative HW limit switch: ",
    "Positive HW limit switch: ",
    "SW limit reached: ",
    "Following Error: ",
    "Drive fault: ",
    "Encoder error: ",
]

SPECIFIC_POSITIONS = ["Position ZERO", "Position RETRACTED"]

GUI_SETTINGS = Settings.load("ZONDA GUI")

class TemperatureLog(QWidget):
    """This Widget allows to view the temperature value of all six actuators."""
    def __init__(self, temp: List[str] = None):
        super().__init__()

        self.stripchart = None

        self.temperatures = temp

        # Switch to using white background and black foreground for pyqtgraph stripcharts

        pg.setConfigOption("background", "w")
        pg.setConfigOption("foreground", "k")

        vbox = QVBoxLayout()
        vbox.addWidget(QLabel("Temperature values in C"))
        vbox.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        self.create_temperature_stripchart_widget = self.create_temperature_stripchart()
        vbox.addWidget(self.create_temperature_stripchart_widget)

        self.create_temperature_box_widget = self.create_temperature_box()
        vbox.addWidget(self.create_temperature_box_widget)

        self.setLayout(vbox)

    def create_temperature_box(self):
        vbox = QVBoxLayout()
        vbox.setSpacing(0)

        for box in range(6):
            wbox = QHBoxLayout()
            wbox.setSpacing(0)
            wbox.addWidget(QLabel(f"Temp.{box + 1}: "))
            wbox.setSpacing(0)
            editbox = QLineEdit()
            editbox.setReadOnly(True)
            editbox.setFixedSize(80, 20)
            wbox.setSpacing(0)
            wbox.addWidget(editbox)
            wbox.setSpacing(0)
            vbox.addLayout(wbox)
            vbox.setSpacing(0)

        create_temperature_box = QGroupBox()
        create_temperature_box.setLayout(vbox)

        return create_temperature_box

    def create_temperature_stripchart(self):
        self.stripchart = StripChart(
            labels={"left": ("measure", "C"), "bottom": ("Time", "d hh:mm:ss")}
        )
        self.stripchart.setInterval(60 * 60 * 12)  # 12h of data
        self.stripchart.set_yrange(0, 40)

        vbox = QVBoxLayout()
        vbox.addStretch(1)
        vbox.addWidget(self.stripchart)

        create_temperature_stripchart = QGroupBox()
        create_temperature_stripchart.setLayout(vbox)

        return create_temperature_stripchart

class ZondaUIView(HexapodUIView):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Hexapod ZONDA Controller")

        self.actuator_states = ActuatorStates(labels=ACTUATOR_STATE_LABELS)

        self.temperature_log = TemperatureLog()

        self.temperature_values = self.temperature_log.create_temperature_box_widget
        self.temperature_values = self.temperature_values.findChildren(QLineEdit)
        for temp in range(len(self.temperature_values)):
            self.temperature_values[temp].setText("0.00")

        self.temperature_stripchart = self.temperature_log.create_temperature_stripchart_widget
        self.temperature_stripchart = self.temperature_stripchart.findChildren(StripChart)
        self.temperature_stripchart = self.temperature_stripchart[0]

        self.init_gui()

    def init_gui(self):

        # The main frame in which all the other frames are located, the outer Application frame

        app_frame = QFrame()
        app_frame.setObjectName("AppFrame")

        # The left part which shows the states and positions

        status_frame = QFrame()
        status_frame.setObjectName("StatusFrame")

        # The right part which has tabs that allow settings, movements, maintenance etc.

        tabs_frame = QFrame()
        tabs_frame.setObjectName("TabsFrame")

        # The states of the Hexapod (contains all the leds)

        states_frame = QFrame()
        states_frame.setObjectName("StatesFrame")

        # The user, machine positions and actuator lengths

        positions_frame = QFrame()
        positions_frame.setObjectName("PositionsFrame")

        hbox = QHBoxLayout()
        vbox_left = QVBoxLayout()
        vbox_right = QVBoxLayout()

        self.create_toolbar()
        self.create_status_bar()

        self.states = States(STATUS_LEDS)

        user_positions_widget = self.create_user_position_widget()
        mach_positions_widget = self.create_machine_position_widget()
        actuator_length_widget = self.create_actuator_length_widget()

        vbox_right.addWidget(user_positions_widget)
        vbox_right.addWidget(mach_positions_widget)
        vbox_right.addWidget(actuator_length_widget)

        positions_frame.setLayout(vbox_right)

        vbox_left.addWidget(self.states)

        states_frame.setLayout(vbox_left)

        hbox.addWidget(states_frame)
        hbox.addWidget(positions_frame)

        status_frame.setLayout(hbox)

        tabbed_widget = self.create_tabbed_widget()

        hbox = QHBoxLayout()
        hbox.addWidget(tabbed_widget)
        tabs_frame.setLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(status_frame)
        hbox.addWidget(tabs_frame)

        app_frame.setLayout(hbox)

        self.setCentralWidget(app_frame)

    def update_status_bar(self, message=None, mode=None, timeout=2000):
        if message:
            self.statusBar().showMessage(message, msecs=timeout)
        if mode:
            self.mode_label.setStyleSheet(
                f"border: 0; " f"color: {'red' if 'Simulator' in mode else 'black'};"
            )

            self.mode_label.setText(f"mode: {mode}")
        self.statusBar().repaint()

    def updatePositions(self, userPositions, machinePositions, actuatorLengths):

        if userPositions is None:
            MODULE_LOGGER.warning("no userPositions passed into updatePositions(), returning.")
            return

        for upos in range(len(self.user_positions)):
            try:
                self.user_positions[upos][1].setText(f"{userPositions[upos]:10.4f}")
            except IndexError:
                MODULE_LOGGER.error(f"IndexError in user_positions, upos = {upos}")

        if machinePositions is None:
            MODULE_LOGGER.warning("no machinePositions passed into updatePositions(), returning.")
            return

        for mpos in range(len(self.mach_positions)):
            self.mach_positions[mpos][1].setText(f"{machinePositions[mpos]:10.4f}")

        if actuatorLengths is None:
            MODULE_LOGGER.warning("no actuatorLengths passed into updatePositions(), returning.")
            return

        for idx, alen in enumerate(self.actuator_lengths):
            alen[1].setText(f"{actuatorLengths[idx]:10.4f}")

    def updateStates(self, states):

        if states is None:
            return

        self.updateControlButton(states[CONTROL_ONOFF])
        self.states.set_states(states)

    def updateControlButton(self, flag):

        self.control.set_selected(on=flag)

    def updateTemperature(self, temp):
        if temp is None:
            MODULE_LOGGER.warning("no temperature passed into updateTemperature(), returning.")
            return
        else:
            #TODO: How to add the 6 temperature values to the stripchart?
            value = temp[0]
            self.temperature_stripchart.update(QDateTime.currentMSecsSinceEpoch(), value)
            for t in range(len(self.temperature_values)):
                self.temperature_values[t].setText(f"{temp[t]:10.4f}")


class ZondaUIModel(HexapodUIModel):
    def __init__(self, connection_type):

        if connection_type == "proxy":
            device = ZondaProxy()
        elif connection_type == "direct":
            device = ZondaController()
            device.connect()
        elif connection_type == "simulator":
            device = ZondaSimulator()
        else:
            raise ValueError(
                f"Unknown type of Hexapod implementation passed into the model: {connection_type}"
            )

        super().__init__(connection_type, device)

        if device is not None:
            MODULE_LOGGER.debug(f"Hexapod initialized as {device.__class__.__name__}")

    def get_speed(self):
        speed_settings = self.device.get_speed()
        return speed_settings["vt"], speed_settings["vr"]

    def get_temperature(self):
        temp = self.device.get_temperature()
        return temp


class ZondaUIController(HexapodUIController):
    def __init__(self, model: ZondaUIModel, view: ZondaUIView):
        super().__init__(model, view)

    def update_values(self):

        super().update_values()

        # Add here any updates to ZONDA specific widgets

        temp = self.model.get_temperature()
        self.view.updateTemperature(temp)


def parse_arguments():
    """
    Prepare the arguments that are specific for this application.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--type",
        dest="type",
        action="store",
        choices={"proxy", "simulator", "direct"},
        help="Specify Hexapod implementation you want to connect to.",
        default="proxy",
    )
    parser.add_argument(
        "--profile",
        default=False,
        action="store_true",
        help="Enable info logging messages with method profile information.",
    )
    return parser.parse_args()


def main():
    lock_file = QLockFile(str(Path("~/zonda_ui.app.lock").expanduser()))

    styles_location = get_resource(":/styles/default.qss")
    app_logo = get_resource(":/icons/logo-zonda.svg")

    args = list(sys.argv)
    args[1:1] = ["-stylesheet", str(styles_location)]
    app = QApplication(args)
    app.setWindowIcon(QIcon(str(app_logo)))

    if lock_file.tryLock(100):

        process_status = ProcessStatus()

        timer_thread = threading.Thread(target=do_every, args=(10, process_status.update))
        timer_thread.daemon = True
        timer_thread.start()

        start_http_server(GUI_SETTINGS.METRICS_PORT)

        args = parse_arguments()

        if args.profile:
            Settings.set_profiling(True)

        if args.type == "proxy":
            proxy = ZondaProxy()
            if not proxy.ping():
                description = "Could not connect to Hexapod Control Server"
                info_text = (
                    "The GUI will start, but the connection button will show a disconnected state. "
                    "Please check if the Control Server is running and start the server if needed. "
                    "Otherwise, check if the correct HOSTNAME for the control server is set in the "
                    "Settings.yaml "
                    "configuration file."
                )

                show_warning_message(description, info_text)

        view = ZondaUIView()
        model = ZondaUIModel(args.type)
        ZondaUIController(model, view)

        view.show()

        return app.exec_()
    else:
        error_message = QMessageBox()
        error_message.setIcon(QMessageBox.Warning)
        error_message.setWindowTitle("Error")
        error_message.setText("The Zonda GUI application is already running!")
        error_message.setStandardButtons(QMessageBox.Ok)

        return error_message.exec()


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(main())
